from PIL import Image
from .. import objects
from .. import NVector, Color
from ..utils.color import from_uint8


class Polygen:
    def __init__(self, x, y):
        self.vertices = [
            NVector(x, y),
            NVector(x+1, y),
            NVector(x+1, y+1),
            NVector(x, y+1),
        ]
        self._has_x = False
        self._has_y = False

    def add_pixel_x(self, x, y):
        i = self.vertices.index(NVector(x, y))
        if len(self.vertices) > i and self.vertices[i+1] == NVector(x, y+1):
            self._has_x = True
            self.vertices.insert(i+1, NVector(x+1, y))
            self.vertices.insert(i+2, NVector(x+1, y+1))
        else:
            raise ValueError()

    def add_pixel_x_neg(self, x, y):
        i = self.vertices.index(NVector(x+1, y))
        if i > 0 and self.vertices[i-1] == NVector(x+1, y+1):
            self._has_x = True
            self.vertices.insert(i, NVector(x, y))
            self.vertices.insert(i, NVector(x, y+1))
        else:
            raise ValueError()

    def add_pixel_y(self, x, y):
        i = self.vertices.index(NVector(x, y))
        if i > 0 and self.vertices[i-1] == NVector(x+1, y):
            self._has_y = True
            if i > 1 and self.vertices[i-2] == NVector(x+1, y+1):
                self.vertices[i-1] = NVector(x, y+1)
            else:
                self.vertices.insert(i, NVector(x, y+1))
                self.vertices.insert(i, NVector(x+1, y+1))
        else:
            raise ValueError()

    def _to_rect(self, id1, id2, scale):
        p1 = self.vertices[id1] * scale
        p2 = self.vertices[id2] * scale
        return objects.Rect((p1+p2)/2, p2-p1)

    def to_shape(self, scale):
        if not self._has_x or not self._has_y:
            return self._to_rect(0, int(len(self.vertices)/2), scale)
        bez = objects.Bezier()
        bez.closed = True
        for point in self.vertices:
            point = point * scale
            if len(bez.vertices) > 1 and (
                bez.vertices[-1].x == bez.vertices[-2].x == point.x or
                bez.vertices[-1].y == bez.vertices[-2].y == point.y
            ):
                bez.vertices[-1] = point
            else:
                bez.add_point(point)

        if len(bez.vertices) > 2 and bez.vertices[0].x == bez.vertices[-1].x == bez.vertices[-2].x:
            bez.vertices.pop()
            bez.out_tangents.pop()
            bez.in_tangents.pop()
        return objects.Path(bez)


def pixel_to_layer_paths(raster, scale=1, stroke_width=None):
    layer = objects.ShapeLayer()
    groups = {}
    processed = set()
    xneg_candidates = set()
    if stroke_width is None:
        stroke_width = 0.1 * scale

    def avail(x, y):
        rid = (x, y)
        return not (
            x < 0 or x >= raster.width or y >= raster.height or
            rid in processed or raster.getpixel(rid) != colort
        )

    def recurse(gen, x, y, xneg):
        processed.add((x, y))
        if avail(x+1, y):
            gen.add_pixel_x(x+1, y)
            recurse(gen, x+1, y, False)
        if avail(x, y+1):
            gen.add_pixel_y(x, y+1)
            recurse(gen, x, y+1, True)
        if xneg and avail(x-1, y):
            xneg_candidates.add((x-1, y))

    for y in range(raster.height):
        for x in range(raster.width):
            pid = (x, y)
            colort = raster.getpixel(pid)
            if colort[-1] == 0 or pid in processed:
                continue

            gen = Polygen(x, y)
            xneg_candidates = set()
            recurse(gen, x, y, False)
            xneg_candidates -= processed
            while xneg_candidates:
                p = next(iter(sorted(xneg_candidates, key=lambda t: (t[1], t[0]))))
                gen.add_pixel_x_neg(*p)
                recurse(gen, p[0], p[1], True)
                processed.add(p)
                xneg_candidates -= processed

            g = groups.setdefault(colort, set())
            g.add(gen.to_shape(scale))

    # Debug
    #for colort, rects in groups.items():
        #for rect in rects:
            #g = layer.add_shape(objects.Group())
            #g.shapes = [rect] + g.shapes
            #g.name = "".join("%02x" % c for c in colort)
            #color = from_uint8(*colort[:3])
            #opacity = colort[-1] / 255 * 100
            #stroke = g.add_shape(objects.Stroke(NVector(0, 0, 0), stroke_width))
            #fill = g.add_shape(objects.Fill())
            #fill.color.value = color
            #fill.opacity.value = 20

    for colort, rects in groups.items():
        g = layer.add_shape(objects.Group())
        g.shapes = list(rects) + g.shapes
        g.name = "".join("%02x" % c for c in colort)
        color = from_uint8(*colort[:3])
        opacity = colort[-1] / 255 * 100
        fill = g.add_shape(objects.Fill())
        fill.color.value = color
        fill.opacity.value = opacity
        if stroke_width > 0:
            stroke = g.add_shape(objects.Stroke(color, stroke_width))
            stroke.opacity.value = opacity

    return layer


def pixel_add_layer_paths(animation, raster):
    return animation.add_layer(pixel_to_layer_paths(raster))


def pixel_add_layer_rects(animation, raster):
    layer = animation.add_layer(objects.ShapeLayer())
    last_rects = {}
    groups = {}

    def merge_up():
        if last_rect and last_rect._start in last_rects:
            yrect = last_rects[last_rect._start]
            if yrect.size.value.x == last_rect.size.value.x and yrect._color == last_rect._color:
                groups[last_rect._color].remove(last_rect)
                yrect.position.value.y += 0.5
                yrect.size.value.y += 1
                rects[last_rect._start] = yrect

    def group(colort):
        return groups.setdefault(colort, set())

    for y in range(raster.height):
        rects = {}
        last_color = None
        last_rect = None
        for x in range(raster.width):
            colort = raster.getpixel((x, y))
            if colort[-1] == 0:
                last_color = 0
                continue
            yrect = last_rects.get(x, None)
            if colort == last_color:
                last_rect.position.value.x += 0.5
                last_rect.size.value.x += 1
            elif yrect and colort == yrect._color and yrect.size.value.x == 1:
                yrect.position.value.y += 0.5
                yrect.size.value.y += 1
                rects[x] = yrect
                last_color = last_rect = colort = None
            else:
                merge_up()
                g = group(colort)
                last_rect = objects.Rect()
                g.add(last_rect)
                last_rect.size.value = NVector(1, 1)
                last_rect.position.value = NVector(x + 0.5, y + 0.5)
                rects[x] = last_rect
                last_rect._start = x
                last_rect._color = colort
            last_color = colort
        merge_up()
        last_rects = rects

    for colort, rects in groups.items():
        g = layer.add_shape(objects.Group())
        g.shapes = list(rects) + g.shapes
        g.name = "".join("%02x" % c for c in colort)
        fill = g.add_shape(objects.Fill())
        fill.color.value = from_uint8(*colort[:3])
        fill.opacity.value = colort[-1] / 255 * 100
        stroke = g.add_shape(objects.Stroke(fill.color.value, 0.1))
        stroke.opacity.value = fill.opacity.value
    return layer


def _vectorizing_func(filenames, frame_delay, framerate, callback):
    if not isinstance(filenames, list):
        filenames = [filenames]

    animation = objects.Animation(0, framerate)
    nframes = 0
    time = 0

    for filename in filenames:
        if isinstance(filename, Image.Image):
            raster = filename
        else:
            raster = Image.open(filename)

        if nframes == 0:
            animation.width = raster.width
            animation.height = raster.height

        if not hasattr(raster, "is_animated"):
            raster.n_frames = 1
            raster.seek = lambda x: None

        for frame in range(raster.n_frames):
            raster.seek(frame)
            new_im = Image.new("RGBA", raster.size)
            new_im.paste(raster)
            duration = frame_delay
            image_duration = raster.info.get("duration", 0)
            if image_duration:
                duration = framerate * image_duration / 1000
            callback(animation, new_im, nframes + frame, time, duration)
            time += duration
            new_im.close()
        nframes += raster.n_frames

    animation.out_point = time
    return animation


def raster_to_embedded_assets(filenames, frame_delay=1, framerate=60, embed_format=None):
    """!
    @brief Loads external assets
    """
    def callback(animation, raster, frame, time, duration):
        asset = objects.assets.Image.embedded(raster, embed_format)
        animation.assets.append(asset)
        layer = animation.add_layer(objects.ImageLayer(asset.id))
        layer.in_point = time
        layer.out_point = layer.in_point + duration

    return _vectorizing_func(filenames, frame_delay, framerate, callback)


def raster_to_linked_assets(filenames, frame_delay=1, framerate=60):
    """!
    @brief Loads external assets
    """
    animation = objects.Animation(frame_delay * len(filenames), framerate)

    for frame, filename in enumerate(filenames):
        asset = objects.assets.Image.linked(filename)
        animation.assets.append(asset)
        layer = animation.add_layer(objects.ImageLayer(asset.id))
        layer.in_point = frame * frame_delay
        layer.out_point = layer.in_point + frame_delay

    return animation


def pixel_to_animation(filenames, frame_delay=1, framerate=60):
    """!
    @brief Converts pixel art to vector
    """
    def callback(animation, raster, frame, time, duration):
        layer = pixel_add_layer_rects(animation, raster.convert("RGBA"))
        layer.in_point = time
        layer.out_point = layer.in_point + duration

    return _vectorizing_func(filenames, frame_delay, framerate, callback)


def pixel_to_animation_paths(filenames, frame_delay=1, framerate=60):
    """!
    @brief Converts pixel art to vector paths

    Slower and yields larger files compared to pixel_to_animation,
    but it produces a single shape for each area with the same color.
    Mostly useful when you want to add your own animations to the loaded image
    """
    def callback(animation, raster, frame, time, duration):
        layer = pixel_add_layer_paths(animation, raster.convert("RGBA"))
        layer.in_point = time
        layer.out_point = layer.in_point + duration

    return _vectorizing_func(filenames, frame_delay, framerate, callback)



from xml.dom import minidom
import enum
from lottie.parsers.sif.sif.core import TypeDescriptor, ObjectRegistry, SifNodeMeta, FrameTime
from lottie.parsers.sif.xml.utils import xml_child_elements, xml_first_element_child


class SifAstNode:
    _subclasses = None
    _tag = None

    @staticmethod
    def from_dom(xml: minidom.Element, param: TypeDescriptor, registry: ObjectRegistry):
        if xml.tagName == param.typename:
            return SifValue.from_dom(xml, param, registry)

        xmltype = xml.getAttribute("type")
        if xmltype != param.typename and xmltype != "weighted_" + param.typename:
            raise ValueError("Invalid type %s (should be %s)" % (xmltype, param.typename))

        return SifAstNode.ast_node_types()[xml.tagName].from_dom(xml, param, registry)

    @staticmethod
    def ast_node_types():
        if SifAstNode._subclasses is None:
            from . import nodes
            SifAstNode._subclasses = {}
            SifAstNode._gather_ast_types(SifAstNode)
        return SifAstNode._subclasses

    @staticmethod
    def _gather_ast_types(cls):
        for subcls in cls.__subclasses__():
            if subcls._tag:
                SifAstNode._subclasses[subcls._tag] = subcls
            SifAstNode._gather_ast_types(subcls)

    def _prepare_to_dom(self, dom: minidom.Document, param: TypeDescriptor):
        element = dom.createElement(self._tag)
        element.setAttribute("type", param.typename)
        return element


class SifValue(SifAstNode):
    def __init__(self, value=None):
        self.value = value

    def __repr__(self):
        return "<%s %r>" % (self.__class__.__name__, self.value)

    @classmethod
    def from_dom(cls, xml: minidom.Element, param: TypeDescriptor, registry: ObjectRegistry):
        return SifValue(param.value_from_xml_element(xml, registry))

    def to_dom(self, dom: minidom.Document, param: TypeDescriptor):
        return param.value_to_xml_element(self.value, dom)


class Interpolation(enum.Enum):
    Auto = "auto"
    Linear = "linear"
    Clamped = "clamped"
    Ease = "halt"
    Constant = "constant"


class SifKeyframe:
    def __init__(self, value, time: FrameTime, before=Interpolation.Clamped, after=Interpolation.Clamped):
        self.value = value
        self.time = time
        self.before = before
        self.after = after

    @classmethod
    def from_dom(cls, xml: minidom.Element, param: TypeDescriptor, registry: ObjectRegistry):
        return cls(
            param.value_from_xml_element(xml_first_element_child(xml), registry),
            FrameTime.parse_string(xml.getAttribute("time"), registry),
            Interpolation(xml.getAttribute("before")),
            Interpolation(xml.getAttribute("after"))
        )

    def to_dom(self, dom: minidom.Document, param: TypeDescriptor):
        element = dom.createElement("waypoint")
        element.setAttribute("time", str(self.time))
        element.setAttribute("before", self.before.value)
        element.setAttribute("after", self.after.value)
        element.appendChild(param.value_to_xml_element(self.value, dom))
        return element

    def __repr__(self):
        return "<SifKeyframe %s %s>" % (self.time, self.value)


class SifAnimated(SifAstNode):
    _tag = "animated"

    def __init__(self):
        self.keyframes = []

    @classmethod
    def from_dom(cls, xml: minidom.Element, param: TypeDescriptor, registry: ObjectRegistry):
        obj = SifAnimated()
        for waypoint in xml_child_elements(xml, "waypoint"):
            obj.keyframes.append(SifKeyframe.from_dom(waypoint, param, registry))
        return obj

    def to_dom(self, dom: minidom.Document, param: TypeDescriptor):
        element = self._prepare_to_dom(dom, param)
        for kf in self.keyframes:
            element.appendChild(kf.to_dom(dom, param))
        return element

    def add_keyframe(self, *args, **kwargs):
        if not kwargs and len(args) == 1 and isinstance(args[0], SifKeyframe):
            keyframe = args[0]
        else:
            keyframe = SifKeyframe(*args, **kwargs)

        self.keyframes.append(keyframe)
        return keyframe



#!/usr/bin/env python3

import sys
import os
import re
import argparse
sys.path.insert(0, os.path.join(
    os.path.dirname(os.path.dirname(os.path.abspath(__file__))),
    "lib"
))
from lottie.parsers.svg.importer import parse_color
from lottie import NVector
from lottie import __version__


def _parse_color(color):
    # Inkscape colors
    if re.match(r"^[0-9a-fA-F]{8}$", color):
        return NVector(*(
            int(color[i:i+2], 16) / 0xff
            for i in range(0, 8, 2)
        ))
    return parse_color(color)



parser = argparse.ArgumentParser(
    description="Converts a CSS color into a normalized array, as used in lottie"
)
parser.add_argument("--version", "-v", action="version", version="%(prog)s - python-lottie " + __version__)
parser.add_argument(
    "color",
    help="Color to inspect (in one of the CSS color formats)"
)

if __name__ == "__main__":
    ns = parser.parse_args()
    print(_parse_color(ns.color).components)



import json
import tempfile
import subprocess
from .base import importer
from ..objects import Animation


@importer("Python script", ["py"])
def import_python_script(file, *a, **kw):

    out = subprocess.check_output(["python", file, "--version"])
    if b"python-lottie script" not in out:
        raise Exception("Not a valid script")

    data = subprocess.check_output(["python", file, "--path", "", "--name", "-", "--format", "json"])
    return Animation.load(json.loads(data))



import math
import typing
import random

from dataclasses import dataclass
from .. import objects
from .color import Color
from ..nvector import NVector
from ..objects import easing
from . import ellipse


class Datum:
    def __init__(self, value: float, color: Color):
        self.value = value
        self.color = color

    def add_style(self, group: objects.Group):
        group.add_shape(objects.Fill(self.color))


class DataSet:
    def __init__(self, data: typing.List[Datum]):
        self.data = data

    def normalize_values(self, sum_to_one=False):
        """!
        Ensures all values are in [0, 1]

        @param sum_to_one if \b True, the sum all values will be 1
        """
        if sum_to_one:
            max_value = sum(datum.value for datum in self.data)
        else:
            max_value = max(datum.value for datum in self.data)

        if max_value == 0:
            return

        for datum in self.data:
            datum.value /= max_value

    def add(self, value: float, color: Color):
        self.data.append(Datum(value, color))


@dataclass
class AnimationSettings:
    #! Time at which the animation starts
    start: float = 0
    #! Duration from start until an item is fully visible
    fade_in: float = 0
    #! Duration the item is fully visible
    stay: float = 0
    #! Duration of the item disappearing
    fade_out: float = 0
    #! End time
    end: float = 0
    #! Offset between items
    offset: float = 0

    @classmethod
    def computed(cls, item_count, duration, stay, full_stay, start=0):
        """!
        @param item_count   Number of items to animate
        @param dutation     Total duration
        @param stay         Duration for which each item is fully visible
        @param full_stay    Duration for all each items are fully visible
        @param start        Start time
        """

        overlap = stay - full_stay
        offset = overlap / (item_count - 1)
        partial_stay = stay + overlap
        inout = duration - partial_stay

        return AnimationSettings(
            start=start,
            fade_in=inout * 2 / 3,
            stay=stay,
            fade_out=inout / 3,
            offset=offset,
            end=start+duration,
        )

    def to_global_times(self):
        """!
        Converts durations into global times
        """
        self.fade_in += self.start
        self.stay += self.fade_in
        self.fade_out += self.stay


class DatumAnimation:
    def __init__(self, datum: Datum, index_in: int, index_out: int, index_count: int, settings: AnimationSettings):
        self.datum = datum
        self.index_in = index_in
        self.index_out = index_out
        off1 = self.index_in * settings.offset
        off2 = self.index_out * settings.offset
        off_max = index_count * settings.offset
        self.times = AnimationSettings(
            start=settings.start + off1,
            end=settings.end - off_max + off2
        )
        self.times.fade_in = self.times.start + settings.fade_in
        self.times.fade_out = self.times.end - settings.fade_out
        self.times.stay = self.times.fade_out


class ChartType:
    def animate(self, area: objects.BoundingBox, anim: DatumAnimation, index: int, total: int, sum_before: float):
        raise NotImplementedError()


class Histogram(ChartType):
    def animate(self, area: objects.BoundingBox, anim: DatumAnimation, index: int, total: int, sum_before: float):
        x = area.width / total * (index + .5) + area.x1
        width = area.width / (total + 1)
        height = area.height * anim.datum.value

        group = objects.Group()
        rect = group.add_shape(objects.Rect())
        anim.datum.add_style(group)

        rect.position.add_keyframe(anim.times.start,    NVector(x, area.y2), easing.EaseOut())
        rect.position.add_keyframe(anim.times.fade_in,  NVector(x, area.y2 - height / 2), easing.EaseOut())
        rect.position.add_keyframe(anim.times.fade_out, NVector(x, area.y2 - height / 2), easing.EaseOut())
        rect.position.add_keyframe(anim.times.end,      NVector(x, area.y2), easing.EaseOut())

        rect.size.add_keyframe(anim.times.start,    NVector(width, 0), easing.EaseOut())
        rect.size.add_keyframe(anim.times.fade_in,  NVector(width, height), easing.EaseOut())
        rect.size.add_keyframe(anim.times.fade_out, NVector(width, height), easing.EaseOut())
        rect.size.add_keyframe(anim.times.end,      NVector(width, 0), easing.EaseOut())

        return group


class PieFan(ChartType):
    def animate(self, area: objects.BoundingBox, anim: DatumAnimation, index: int, total: int, sum_before: float):
        angle_start = sum_before * math.pi * 2
        angle_delta = anim.datum.value * math.pi * 2

        rad = min(area.width, area.height) / 2
        ellipser = ellipse.Ellipse(area.center(), NVector(rad, rad), 0)

        group = objects.Group()
        shape = group.add_shape(objects.Path())

        cache = {}

        self.add_transition(shape, ellipser, angle_start, 0, angle_delta, anim.times.start, anim.times.fade_in, cache)
        self.add_transition(shape, ellipser, angle_start, angle_delta, 0, anim.times.fade_out, anim.times.end, cache)

        anim.datum.add_style(group)
        return group

    def add_arc(self, shape, ellipser, angle_start, angle_delta, time, cache):
        if angle_delta not in cache:
            if angle_delta == 0:
                bez = objects.Bezier()
                p = ellipser.center + NVector(
                    math.cos(angle_start) * ellipser.radii.x,
                    math.sin(angle_start) * ellipser.radii.x
                )
                bez.add_point(p)
                bez.add_point(p)
                bez.add_point(p)
                bez.add_point(p)
            else:
                bez = ellipser.to_bezier(angle_start, angle_delta, angle_delta / 4)
                bez.in_tangents[0] = bez.out_tangents[-1] = NVector(0, 0)

            bez.add_point(ellipser.center)
            bez.add_point(ellipser.center)
            bez.close()

            cache[angle_delta] = bez
        else:
            bez = cache[angle_delta]

        shape.shape.add_keyframe(time, bez)

    def add_transition(self, shape, ellipser, angle_start, angle_delta_from, angle_delta_to, time_from, time_to, cache):
        chunks = round(max(angle_delta_from, angle_delta_to) / math.pi * 100)
        for step in range(chunks + 1):
            f = step / chunks
            angle_delta = f * angle_delta_to + (1-f) * angle_delta_from
            time = f * time_to + (1-f) * time_from
            self.add_arc(shape, ellipser, angle_start, angle_delta, time, cache)


class PieGrow(ChartType):
    def animate(self, area: objects.BoundingBox, anim: DatumAnimation, index: int, total: int, sum_before: float):
        angle_start = sum_before * math.pi * 2
        angle_delta = anim.datum.value * math.pi * 2

        group = objects.Group()
        shape = group.add_shape(objects.Path())

        rad = min(area.width, area.height) / 2
        bez_0 = self.arc_to_bezier(area.center(), 0, 0, 0)
        bez_1 = self.arc_to_bezier(area.center(), angle_start, angle_delta, rad)

        shape.shape.add_keyframe(anim.times.start,    bez_0, easing.EaseOut())
        shape.shape.add_keyframe(anim.times.fade_in,  bez_1, easing.EaseOut())
        shape.shape.add_keyframe(anim.times.fade_out, bez_1, easing.EaseOut())
        shape.shape.add_keyframe(anim.times.end,      bez_0, easing.EaseOut())

        anim.datum.add_style(group)
        return group

    def arc_to_bezier(self, center, angle_start, angle_delta, radius):
        if radius == 0:
            bez = objects.Bezier()
            p = center
            bez.add_point(p)
            bez.add_point(p)
            bez.add_point(p)
            bez.add_point(p)
        elif angle_delta == 0:
            bez = objects.Bezier()
            p = center + NVector(math.cos(angle_start) * radius, math.sin(angle_start) * radius)
            bez.add_point(p)
            bez.add_point(p)
            bez.add_point(p)
            bez.add_point(p)
        else:
            ellipser = ellipse.Ellipse(center, NVector(radius, radius), 0)
            bez = ellipser.to_bezier(angle_start, angle_delta, angle_delta / 4)
            bez.in_tangents[0] = bez.out_tangents[-1] = NVector(0, 0)

        # Dunno why we need to add it twice but it doesn't work with just one o_O
        bez.add_point(center)
        bez.add_point(center)
        bez.close()

        return bez


class AnimationOrder:
    def indices(self, data: DataSet):
        """!
        Returns two list of indices indicating the order of animation
        for fade in and out of the elements

        The lists are laid out so that the nth element is the index at
        which that element animates at
        """
        raise NotImplementedError

    def __call__(self, data: DataSet):
        return self.indices(data)


class RandomOrder(AnimationOrder):
    def indices(self, data: DataSet):
        indices_in = list(range(len(data.data)))
        indices_out = list(indices_in)
        random.shuffle(indices_in)
        random.shuffle(indices_out)
        return indices_in, indices_out


class SequentialOrder(AnimationOrder):
    def indices(self, data: DataSet):
        indices = list(range(len(data.data)))
        return indices, indices


class SortValueOrder(AnimationOrder):
    def __init__(self, largest_first=True):
        self.largest_first = largest_first

    def indices(self, data: DataSet):
        info = sorted(enumerate(data.data), key=lambda it: it[1].value)
        if self.largest_first:
            info = reversed(info)

        ranking = [it[0] for it in info]
        indices = list(map(ranking.index, range(len(data.data))))
        return indices, indices


class SimultaneousOrder(AnimationOrder):
    def indices(self, data: DataSet):
        n = len(data.data)
        return [0] * n, [n-1] * n


class FixedOrder(AnimationOrder):
    def __init__(self, indices_in, indices_out=None):
        self.indices_in = indices_in
        self.indices_out = indices_out or indices_in

    def indices(self, data: DataSet):
        return self.indices_in, self.indices_out


class Chart:
    def __init__(self, type=Histogram(), area=None, data=None, animation=None, order=SimultaneousOrder()):
        self.type = type
        self.area = area or objects.BoundingBox(0, 0, 512, 512)
        self.timing = animation or AnimationSettings()
        self.order = order

        if isinstance(data, DataSet):
            self.data = data
        else:
            self.data = DataSet(data or [])

    def compute_animation(self, *a, **kw):
        self.timing = AnimationSettings.computed(len(self.data.data), *a, **kw)

    def shapes(self):
        total = len(self.data.data)
        indices_in, indices_out = self.order(self.data)
        shapes = []

        value_sum = 0

        for i, (datum, ii, io) in enumerate(zip(self.data.data, indices_in, indices_out)):
            anim = DatumAnimation(datum, ii, io, total, self.timing)
            shapes.append(self.type.animate(self.area, anim, i, total, value_sum))
            value_sum += datum.value

        return shapes

    def layer(self):
        layer = objects.ShapeLayer()
        layer.shapes = self.shapes()
        return layer

    def animation(self):
        animation = objects.Animation()
        animation.width = self.area.x2
        animation.height = self.area.y2
        animation.in_point = self.timing.start
        animation.out_point = self.timing.end
        animation.add_layer(self.layer())
        return animation



import math
from ... import objects
from ...objects import easing
from . import api, ast
from ... import NVector, PolarVector

try:
    from ...utils import font
    has_font = True
except ImportError:
    has_font = False


def convert(canvas: api.Canvas):
    return Converter().convert(canvas)


class Converter:
    def __init__(self):
        pass

    def _animated(self, sifval):
        return isinstance(sifval, ast.SifAnimated)

    def convert(self, canvas: api.Canvas):
        self.canvas = canvas
        self.animation = objects.Animation(
            self._time(canvas.end_time),
            canvas.fps
        )
        self.animation.in_point = self._time(canvas.begin_time)
        self.animation.width = canvas.width
        self.animation.height = canvas.height
        self.view_p1 = NVector(canvas.view_box[0], canvas.view_box[1])
        self.view_p2 = NVector(canvas.view_box[2], canvas.view_box[3])
        self.target_size = NVector(canvas.width, canvas.height)
        self.shape_layer = self.animation.add_layer(objects.ShapeLayer())
        self.gamma = NVector(canvas.gamma_r, canvas.gamma_g, canvas.gamma_b)
        self._process_layers(canvas.layers, self.shape_layer)
        return self.animation

    def _time(self, t: api.FrameTime):
        return self.canvas.time_to_frames(t)

    def _process_layers(self, layers, parent):
        old_gamma = self.gamma

        for layer in reversed(layers):
            if not layer.active:
                continue
            elif isinstance(layer, api.GroupLayerBase):
                parent.add_shape(self._convert_group(layer))
            elif isinstance(layer, api.RectangleLayer):
                parent.add_shape(self._convert_fill(layer, self._convert_rect))
            elif isinstance(layer, api.CircleLayer):
                parent.add_shape(self._convert_fill(layer, self._convert_circle))
            elif isinstance(layer, api.StarLayer):
                parent.add_shape(self._convert_fill(layer, self._convert_star))
            elif isinstance(layer, api.PolygonLayer):
                parent.add_shape(self._convert_fill(layer, self._convert_polygon))
            elif isinstance(layer, api.RegionLayer):
                parent.add_shape(self._convert_fill(layer, self._convert_bline))
            elif isinstance(layer, api.AbstractOutline):
                parent.add_shape(self._convert_outline(layer, self._convert_bline))
            elif isinstance(layer, api.GradientLayer):
                parent.add_shape(self._convert_gradient(layer, parent))
            elif isinstance(layer, api.TransformDown):
                shape = self._convert_transform_down(layer)
                parent.add_shape(shape)
                parent = shape
            elif isinstance(layer, api.TextLayer):
                if has_font:
                    parent.add_shape(self._convert_fill(layer, self._convert_text))
            elif isinstance(layer, api.ColorCorrectLayer):
                self.gamma = self.gamma * NVector(layer.gamma.value, layer.gamma.value, layer.gamma.value)

        self.gamma = old_gamma

    def _convert_group(self, layer: api.GroupLayer):
        shape = objects.Group()
        self._set_name(shape, layer)
        shape.transform.anchor_point = self._adjust_coords(self._convert_vector(layer.origin))
        self._convert_transform(layer.transformation, shape.transform)
        self._process_layers(layer.layers, shape)
        shape.transform.opacity = self._adjust_animated(
            self._convert_scalar(layer.amount),
            lambda x: x*100
        )
        return shape

    def _convert_transform(self, sif_transform: api.AbstractTransform, lottie_transform: objects.Transform):
        if isinstance(sif_transform, api.BoneLinkTransform):
            base_transform = sif_transform.base_value
        else:
            base_transform = sif_transform

        position = self._adjust_coords(self._convert_vector(base_transform.offset))
        rotation = self._adjust_angle(self._convert_scalar(base_transform.angle))
        scale = self._adjust_animated(
            self._convert_vector(base_transform.scale),
            lambda x: x * 100
        )

        lottie_transform.skew_axis = self._adjust_angle(self._convert_scalar(base_transform.skew_angle))

        if isinstance(sif_transform, api.BoneLinkTransform):
            lottie_transform.position = position
            lottie_transform.rotation = rotation
            lottie_transform.scale = scale
            #bone = sif_transform.bone
            #b_pos = self._adjust_coords(self._convert_vector(bone.origin))
            #old_anchor = lottie_transform.anchor_point

            #if sif_transform.translate:
                #self._mix_animations_into(
                    #[position, b_pos, old_anchor],
                    #lottie_transform.position,
                    #lambda base_p, bone_p, anchor: (anchor-self.target_size/2)/2+self.target_size/2
                #)
            #else:
                #lottie_transform.position = position

            #lottie_transform.anchor_point = b_pos
            #lottie_transform.anchor_point.value += NVector(100,0)

            #if sif_transform.rotate:
                #b_rot = self._convert_scalar(bone.angle)
                #self._mix_animations_into([rotation, b_rot], lottie_transform.rotation, lambda a, b: a-b)
            #else:
                #lottie_transform.rotation = rotation

            #if sif_transform.scale_y:
                #b_scale = self._convert_scalar(bone.scalelx)
                #self._mix_animations_into(
                    #scale, b_scale, lottie_transform.scale,
                    #lambda a, b: NVector(a.x, a.y * b)
                #)
            #else:
                #lottie_transform.scale = scale
        else:
            lottie_transform.position = position
            lottie_transform.rotation = rotation
            lottie_transform.scale = scale

    def _mix_animations_into(self, animations, output, mix):
        if not any(x.animated for x in animations):
            output.value = mix(*(x.value for x in animations))
        else:
            for vals in self._mix_animations(*animations):
                time = vals.pop(0)
                output.add_keyframe(time, mix(*vals))

    def _convert_fill(self, layer, converter):
        shape = objects.Group()
        self._set_name(shape, layer)
        shape.add_shape(converter(layer))
        if layer.invert.value:
            shape.add_shape(objects.Rect(self.target_size/2, self.target_size))

        fill = objects.Fill()
        fill.color = self._convert_color(layer.color)
        fill.opacity = self._adjust_animated(
            self._convert_scalar(layer.amount),
            lambda x: x * 100
        )
        shape.add_shape(fill)
        return shape

    def _convert_linecap(self, lc: api.LineCap):
        if lc == api.LineCap.Rounded:
            return objects.LineCap.Round
        if lc == api.LineCap.Squared:
            return objects.LineCap.Square
        return objects.LineCap.Butt

    def _convert_cusp(self, lc: api.CuspStyle):
        if lc == api.CuspStyle.Miter:
            return objects.LineJoin.Miter
        if lc == api.CuspStyle.Bevel:
            return objects.LineJoin.Bevel
        return objects.LineJoin.Round

    def _convert_outline(self, layer: api.AbstractOutline, converter):
        shape = objects.Group()
        self._set_name(shape, layer)
        shape.add_shape(converter(layer))
        stroke = objects.Stroke()
        stroke.color = self._convert_color(layer.color)
        stroke.line_cap = self._convert_linecap(layer.start_tip)
        stroke.line_join = self._convert_cusp(layer.cusp_type)
        stroke.width = self._adjust_scalar(self._convert_scalar(layer.width))
        shape.add_shape(stroke)
        return shape

    def _convert_rect(self, layer: api.RectangleLayer):
        rect = objects.Rect()
        p1 = self._adjust_coords(self._convert_vector(layer.point1))
        p2 = self._adjust_coords(self._convert_vector(layer.point2))
        if p1.animated or p2.animated:
            for time, p1v, p2v in self._mix_animations(p1, p2):
                rect.position.add_keyframe(time, (p1v + p2v) / 2)
                rect.size.add_keyframe(time, abs(p2v - p1v))
            pass
        else:
            rect.position.value = (p1.value + p2.value) / 2
            rect.size.value = abs(p2.value - p1.value)
        rect.rounded = self._adjust_scalar(self._convert_scalar(layer.bevel))
        return rect

    def _convert_circle(self, layer: api.CircleLayer):
        shape = objects.Ellipse()
        shape.position = self._adjust_coords(self._convert_vector(layer.origin))
        radius = self._adjust_scalar(self._convert_scalar(layer.radius))
        shape.size = self._adjust_add_dimension(radius, lambda x: NVector(x, x) * 2)
        return shape

    def _convert_star(self, layer: api.StarLayer):
        shape = objects.Star()
        shape.position = self._adjust_coords(self._convert_vector(layer.origin))
        shape.inner_radius = self._adjust_scalar(self._convert_scalar(layer.radius2))
        shape.outer_radius = self._adjust_scalar(self._convert_scalar(layer.radius1))
        shape.rotation = self._adjust_animated(
            self._convert_scalar(layer.angle),
            lambda x: 90-x
        )
        shape.points = self._convert_scalar(layer.points)
        if layer.regular_polygon.value:
            shape.star_type = objects.StarType.Polygon
        return shape

    def _mix_animations(self, *animatable):
        times = set()
        for v in animatable:
            self._force_animated(v)
            for kf in v.keyframes:
                times.add(kf.time)

        for time in sorted(times):
            yield [time] + [v.get_value(time) for v in animatable]

    def _force_animated(self, lottieval):
        if not lottieval.animated:
            v = lottieval.value
            lottieval.add_keyframe(0, v)
            lottieval.add_keyframe(self.animation.out_point, v)

    def _convert_easing_part(self, interp: api.Interpolation):
        if interp == api.Interpolation.Linear:
            return easing.Linear()
        return easing.Sigmoid()

    def _convert_easing(self, start: api.Interpolation, end: api.Interpolation):
        if api.Interpolation.Constant in (start, end):
            return easing.Hold()
        if start == end:
            return self._convert_easing_part(start)
        return easing.Split(self._convert_easing_part(start), self._convert_easing_part(end))

    def _convert_animatable(self, v: ast.SifAstNode, lot: objects.properties.AnimatableMixin):
        if self._animated(v):
            if len(v.keyframes) == 1:
                lot.value = self._convert_ast_value(v.keyframes[0].value)
            else:
                for i, kf in enumerate(v.keyframes):
                    if i+1 < len(v.keyframes):
                        start = kf.after
                        end = v.keyframes[i+1].before
                        ease = self._convert_easing(start, end)
                    else:
                        ease = easing.Linear()

                    lot.add_keyframe(self._time(kf.time), self._convert_ast_value(kf.value), ease)
        else:
            lot.value = self._convert_ast_value(v)
        return lot

    def _convert_ast_value(self, v):
        if isinstance(v, ast.SifRadialComposite):
            return self._polar(v.radius.value, v.theta.value, 1)
        elif isinstance(v, ast.SifValue):
            return v.value
        elif isinstance(v, ast.SifVectorComposite):
            return NVector(v.x.value, v.y.value)
        else:
            return v

    def _converted_vector_values(self, v):
        if isinstance(v, ast.SifRadialComposite):
            return [self._convert_scalar(v.radius), self._convert_scalar(v.theta)]
        return self._convert_vector(v)

    def _convert_color(self, v: ast.SifAstNode):
        return self._adjust_animated(
            self._convert_animatable(v, objects.ColorValue()),
            self._color_gamma
        )

    def _convert_vector(self, v: ast.SifAstNode):
        return self._convert_animatable(v, objects.MultiDimensional())

    def _convert_scalar(self, v: ast.SifAstNode):
        return self._convert_animatable(v, objects.Value())

    def _color_gamma(self, color):
        color = color.clone()
        for i in range(3):
            color[i] = color[i] ** (1/self.gamma[i])
        return color

    def _adjust_animated(self, lottieval, transform):
        if lottieval.animated:
            for kf in lottieval.keyframes:
                if kf.start is not None:
                    kf.start = transform(kf.start)
                if kf.end is not None:
                    kf.end = transform(kf.end)
        else:
            lottieval.value = transform(lottieval.value)
        return lottieval

    def _adjust_scalar(self, lottieval: objects.Value):
        return self._adjust_animated(lottieval, self._scalar_mult)

    def _adjust_angle(self, lottieval: objects.Value):
        return self._adjust_animated(lottieval, lambda x: -x)

    def _adjust_add_dimension(self, lottieval, transform):
        to_val = objects.MultiDimensional()
        to_val.animated = lottieval.animated
        if lottieval.animated:
            to_val.keyframes = []
            for kf in lottieval.keyframes:
                if kf.start is not None:
                    kf.start = transform(kf.start[0])
                if kf.end is not None:
                    kf.end = transform(kf.end[0])
                to_val.keyframes.append(kf)
        else:
            to_val.value = transform(lottieval.value)
        return to_val

    def _scalar_mult(self, x):
        return x * 60

    def _adjust_coords(self, lottieval: objects.MultiDimensional):
        return self._adjust_animated(lottieval, self._coord)

    def _coord(self, val: NVector):
        return NVector(
            self.target_size.x * (val.x / (self.view_p2.x - self.view_p1.x) + 0.5),
            self.target_size.y * (val.y / (self.view_p2.y - self.view_p1.y) + 0.5),
        )

    def _convert_polygon(self, layer: api.PolygonLayer):
        lot = objects.Path()
        animatables = [self._convert_vector(layer.origin)] + [
            self._convert_vector(p)
            for p in layer.points
        ]
        animated = any(x.animated for x in animatables)
        if not animated:
            lot.shape.value = self._polygon([x.value for x in animatables[1:]], animatables[0].value)
        else:
            for values in self._mix_animations(*animatables):
                time = values[0]
                origin = values[1]
                points = values[2:]
                lot.shape.add_keyframe(time, self._polygon(points, origin))
        return lot

    def _polygon(self, points, origin):
        bezier = objects.Bezier()
        bezier.closed = True
        for point in points:
            bezier.add_point(self._coord(point+origin))
        return bezier

    def _convert_bline(self, layer: api.AbstractOutline):
        lot = objects.Path()
        closed = layer.bline.loop
        animatables = [
            self._convert_vector(layer.origin)
        ]
        for p in layer.bline.points:
            animatables += [
                self._convert_vector(p.point),
                self._convert_scalar(p.t1.radius) if hasattr(p.t1, "radius") else objects.Value(0),
                self._convert_scalar(p.t1.theta) if hasattr(p.t1, "radius") else objects.Value(0),
                self._convert_scalar(p.t2.radius) if hasattr(p.t2, "radius") else objects.Value(0),
                self._convert_scalar(p.t2.theta) if hasattr(p.t2, "radius") else objects.Value(0)
            ]
        animated = any(x.animated for x in animatables)
        if not animated:
            lot.shape.value = self._bezier(
                closed, [x.value for x in animatables[1:]], animatables[0].value, layer.bline.points
            )
        else:
            for values in self._mix_animations(*animatables):
                time = values[0]
                origin = values[1]
                values = values[2:]
                lot.shape.add_keyframe(time, self._bezier(closed, values, origin, layer.bline.points))
        return lot

    def _bezier(self, closed, values, origin, points):
        chunk_size = 5
        bezier = objects.Bezier()
        bezier.closed = closed
        for i in range(0, len(values), chunk_size):
            point, r1, a1, r2, a2 = values[i:i+chunk_size]
            sifvert = point+origin
            vert = self._coord(sifvert)
            if not points[i//chunk_size].split_radius.value:
                r2 = r1
            if not points[i//chunk_size].split_angle.value:
                a2 = a1
            t1 = self._coord(sifvert + self._polar(r1, a1, 1)) - vert
            t2 = self._coord(sifvert + self._polar(r2, a2, 2)) - vert
            bezier.add_point(vert, t1, t2)
        return bezier

    def _polar(self, radius, angle, dir):
        offset_angle = 0
        if dir == 1:
            offset_angle += 180
        return PolarVector(radius/3, (angle+offset_angle) * math.pi / 180)

    def _convert_transform_down(self, tl: api.TransformDown):
        group = objects.Group()
        self._set_name(group, tl)

        if isinstance(tl, api.TranslateLayer):
            group.transform.anchor_point.value = self.target_size / 2
            group.transform.position = self._adjust_coords(self._convert_vector(tl.origin))
        elif isinstance(tl, api.RotateLayer):
            group.transform.anchor_point = self._adjust_coords(self._convert_vector(tl.origin))
            group.transform.position = group.transform.anchor_point.clone()
            group.transform.rotation = self._adjust_angle(self._convert_scalar(tl.amount))
        elif isinstance(tl, api.ScaleLayer):
            group.transform.anchor_point = self._adjust_coords(self._convert_vector(tl.center))
            group.transform.position = group.transform.anchor_point.clone()
            group.transform.scale = self._adjust_add_dimension(
                self._convert_scalar(tl.amount),
                self._zoom_to_scale
            )

        return group

    def _zoom_to_scale(self, value):
        zoom = math.e ** value * 100
        return NVector(zoom, zoom)

    def _set_name(self, lottie, sif):
        lottie.name = sif.desc if sif.desc is not None else sif.__class__.__name__

    def _convert_gradient(self, layer: api.GradientLayer, parent):
        group = objects.Group()

        parent_shapes = parent.shapes
        parent.shapes = []
        if isinstance(parent, objects.Group):
            parent.shapes.append(parent_shapes[-1])

        self._gradient_gather_shapes(parent_shapes, group)

        gradient = objects.GradientFill()
        self._set_name(gradient, layer)
        group.add_shape(gradient)
        gradient.colors = self._convert_gradient_stops(layer.gradient)
        gradient.opacity = self._adjust_animated(
            self._convert_scalar(layer.amount),
            lambda x: x * 100
        )

        if isinstance(layer, api.LinearGradient):
            gradient.start_point = self._adjust_coords(self._convert_vector(layer.p1))
            gradient.end_point = self._adjust_coords(self._convert_vector(layer.p2))
            gradient.gradient_type = objects.GradientType.Linear
        elif isinstance(layer, api.RadialGradient):
            gradient.gradient_type = objects.GradientType.Radial
            gradient.start_point = self._adjust_coords(self._convert_vector(layer.center))
            radius = self._adjust_animated(self._convert_scalar(layer.radius), lambda x: x*45)
            if not radius.animated and not gradient.start_point.animated:
                gradient.end_point.value = gradient.start_point.value + NVector(radius.value, radius.value)
            else:
                for time, c, r in self._mix_animations(gradient.start_point.clone(), radius):
                    gradient.end_point.add_keyframe(time, c + NVector(r + r))

        return group

    def _gradient_gather_shapes(self, shapes, output: objects.Group):
        for shape in shapes:
            if isinstance(shape, objects.Shape):
                output.add_shape(shape)
            elif isinstance(shape, objects.Group):
                self._gradient_gather_shapes(shape.shapes, output)

    def _convert_gradient_stops(self, sif_gradient):
        stops = objects.GradientColors()
        if not self._animated(sif_gradient):
            stops.set_stops(self._flatten_gradient_colors(sif_gradient.value))
            stops.count = len(sif_gradient.value)
        else:
            # TODO easing
            for kf in sif_gradient.keyframes:
                stops.add_keyframe(self._time(kf.time), self._flatten_gradient_colors(kf.value))
                stops.count = len(kf.value)

        return stops

    def _flatten_gradient_colors(self, stops):
        return [
            (stop.pos, self._color_gamma(stop.color))
            for stop in stops
        ]

    def _convert_text(self, layer: api.TextLayer):
        shape = font.FontShape(layer.text.value, font.FontStyle(layer.family.value, 110, font.TextJustify.Center))
        shape.refresh()
        trans = shape.wrapped.transform
        trans.anchor_point.value = shape.wrapped.bounding_box().center()
        trans.anchor_point.value.x /= 2
        trans.position = self._adjust_coords(self._convert_vector(layer.origin))
        trans.scale = self._adjust_animated(
            self._convert_vector(layer.size),
            lambda v: v * 100
        )
        return shape



#!/usr/bin/env python3

import re
import sys
import os
import argparse
sys.path.insert(0, os.path.join(
    os.path.dirname(os.path.dirname(os.path.abspath(__file__))),
    "lib"
))
from lottie.exporters import exporters
from lottie.importers import importers
from lottie.utils.stripper import float_strip, heavy_strip
from lottie import __version__


desc = """Converts between multiple formats

Supported formats:

- Input:
%s

- Output:
%s
""" % (
    "\n".join("%s- %s" % (" "*2, e.name) for e in importers),
    "\n".join("%s- %s" % (" "*2, e.name) for e in exporters),
)

parser = argparse.ArgumentParser(
    description=desc,
    formatter_class=argparse.RawTextHelpFormatter,
    conflict_handler='resolve'
)
parser.add_argument("--version", "-v", action="version", version="%(prog)s - python-lottie " + __version__)

group = importers.set_options(parser)
group.add_argument(
    "infile",
    default="-",
    nargs="?",
    help="Input file"
)
group.add_argument(
    "--input-format", "-if",
    default=None,
    choices=[importer.slug for importer in importers],
    help="Explicit output format (if missing implied by the input filename)",
)


group = exporters.set_options(parser)

group.add_argument(
    "outfile",
    default="-",
    nargs="?",
    help="Output file"
)
group.add_argument(
    "--output-format", "-of",
    default=None,
    choices=[exporter.slug for exporter in exporters],
    help="Explicit output format (if missing implied by the output filename)",
)
group.add_argument(
    "--sanitize",
    default=False,
    action="store_true",
    help="Ensure the animation is 512x512 and 30 or 60 fps",
)
group.add_argument(
    "--optimize",
    "-O",
    default=1,
    type=int,
    choices=[0, 1, 2],
    help="Optimize the animation parameter:\n" +
         " * 0 no optimization\n" +
         " * 1 truncate floats\n" +
         " * 2 truncate floats and names"
)
group.add_argument(
    "--fps",
    default=None,
    type=int,
    help="If present, changes the output fps to match this value"
)
group.add_argument(
    "--width",
    default=None,
    type=int,
    help="Override output width",
)
group.add_argument(
    "--height",
    default=None,
    type=int,
    help="Override output height",
)


def print_dep_message(loader):
    if not loader.failed_modules:
        return

    sys.stderr.write("Make sure you have the correct dependencies installed\n")

    for failed, dep in loader.failed_modules.items():
        sys.stderr.write("For %s install %s\n" % (failed, dep))


if __name__ == "__main__":
    ns = parser.parse_args()
    if ns.infile == "-" and not ns.input_format:
        parser.print_help()

    infile = ns.infile
    importer = None
    if infile == "-":
        infile = sys.stdin
    else:
        suf = os.path.splitext(infile)[1][1:]
        for p in importers:
            if suf in p.extensions:
                importer = p
                break
    if ns.input_format:
        importer = None
        for p in importers:
            if p.slug == ns.input_format:
                importer = p
                break
    if not importer:
        sys.stderr.write("Unknown importer\n")
        print_dep_message(importers)
        sys.exit(1)

    outfile = ns.outfile
    exporter = None
    if outfile == "-":
        outfile = sys.stdout
    else:
        exporter = exporters.get_from_filename(outfile)
    if ns.output_format:
        exporter = exporters.get(ns.output_format)

    if not exporter:
        sys.stderr.write("Unknown exporter\n")
        print_dep_message(exporters)
        sys.exit(1)

    i_options = {}
    for opt in importer.extra_options:
        i_options[opt.name] = getattr(ns, opt.nsvar(importer.slug))

    o_options = exporter.argparse_options(ns)

    an = importer.process(infile, **i_options)

    if ns.fps:
        an.frame_rate = ns.fps

    if ns.width or ns.height:
        width = ns.width
        height = ns.height
        if not width:
            width = an.width * height / an.height
        if not height:
            height = an.height * width / an.width
        an.scale(width, height)

    if ns.optimize == 1:
        float_strip(an)
    elif ns.optimize >= 2:
        heavy_strip(an)

    exporter.process(an, outfile, **o_options)



import io
import os

import cv2
import numpy
from PIL import Image

from .cairo import export_png
from .gif import _log_frame
from .base import exporter
from ..parsers.baseporter import ExtraOption


## @see http://www.fourcc.org/codecs.php
formats4cc = {
    "avi": cv2.VideoWriter_fourcc(*"XVID"),
    "mp4": cv2.VideoWriter_fourcc(*'MP4V'),
    #"mp4": cv2.VideoWriter_fourcc(*'X264'),
    "webm": cv2.VideoWriter_fourcc(*'VP80'),
}


@exporter("Video", list(formats4cc.keys()), [
    ExtraOption("format", default=None, help="Specific video format", choices=list(formats4cc.keys())),
], [], "video")
def export_video(animation, fp, format=None):
    start = int(animation.in_point)
    end = int(animation.out_point)
    if format is None:
        fn = fp if isinstance(fp, str) else fp.name
        format = os.path.splitext(fn)[1][1:]
    fmt = formats4cc[format]
    video = cv2.VideoWriter(fp, fmt, animation.frame_rate, (animation.width, animation.height))

    for i in range(start, end+1):
        _log_frame(format, i, end)
        file = io.BytesIO()
        export_png(animation, file, i)
        file.seek(0)
        video.write(cv2.cvtColor(numpy.array(Image.open(file)), cv2.COLOR_RGB2BGR))

    _log_frame(format)
    video.release()



from .base import LottieObject, LottieProp, PseudoBool
from .properties import Value, MultiDimensional, ColorValue
from ..nvector import NVector
from ..utils.color import Color
from .helpers import VisualObject


## @ingroup Lottie
class EffectValue(VisualObject):
    """!
    Value for an effect
    """
    ## %Effect value type.
    type = None
    _classses = {}

    _props = [
        LottieProp("effect_index", "ix", int, False),
        LottieProp("type", "ty", int, False),
    ]

    def __init__(self):
        super().__init__()
        ## Effect Index. Used for expressions.
        self.effect_index = None

    @classmethod
    def _load_get_class(cls, lottiedict):
        if not EffectValue._classses:
            EffectValue._classses = {
                sc.type: sc
                for sc in EffectValue.__subclasses__()
            }
        return EffectValue._classses[lottiedict["ty"]]

    def __str__(self):
        return self.name or super().__str__()


## @ingroup Lottie
class Effect(VisualObject):
    """!
    Layer effect
    """
    ## %Effect type.
    type = None
    _classses = {}

    _props = [
        LottieProp("effect_index", "ix", int, False),
        LottieProp("type", "ty", int, False),
        LottieProp("effects", "ef", EffectValue, True),
        LottieProp("enabled", "en", bool, False),
    ]
    _effects = []

    def __init__(self, *args, **kwargs):
        super().__init__()
        ## Effect Index. Used for expressions.
        self.effect_index = None
        ## Effect parameters
        self.effects = self._load_values(*args, **kwargs)
        ## Whether the effect is enabled
        self.enabled = None

    @classmethod
    def _load_get_class(cls, lottiedict):
        if not Effect._classses:
            Effect._classses = {
                sc.type: sc
                for sc in Effect.__subclasses__()
            }
        type = lottiedict["ty"]

        if type in Effect._classses:
            return Effect._classses[type]
        else:
            return Effect

    def _load_values(self, *args, **kwargs):
        values = []
        for i, (name, type) in enumerate(self._effects):
            val = []
            if len(args) > i:
                val = [args[i]]
            if name in kwargs:
                val = [kwargs[name]]
            values.append(type(*val))
        return values

    def __getattr__(self, key):
        for i, (name, type) in enumerate(self._effects):
            if name == key:
                return self.effects[i].value
        raise AttributeError(key)

    def __str__(self):
        return self.name or super().__str__()


## @ingroup Lottie
## @ingroup LottieCheck
class EffectNoValue(EffectValue):
    _props = []


## @ingroup Lottie
class EffectValueSlider(EffectValue):
    _props = [
        LottieProp("value", "v", Value, False),
    ]
    ## %Effect type.
    type = 0

    def __init__(self, value=0):
        EffectValue.__init__(self)
        ## Effect value.
        self.value = Value(value)


## @ingroup Lottie
class EffectValueAngle(EffectValue):
    _props = [
        LottieProp("value", "v", Value, False),
    ]
    ## %Effect type.
    type = 1

    def __init__(self, angle=0):
        EffectValue.__init__(self)
        ## Effect value.
        self.value = Value(angle)


## @ingroup Lottie
class EffectValueColor(EffectValue):
    _props = [
        LottieProp("value", "v", ColorValue, False),
    ]
    ## %Effect type.
    type = 2

    def __init__(self, value=Color(0, 0, 0)):
        EffectValue.__init__(self)
        ## Effect value.
        self.value = ColorValue(value)


## @ingroup Lottie
class EffectValuePoint(EffectValue):
    _props = [
        LottieProp("value", "v", MultiDimensional, False),
    ]
    ## %Effect type.
    type = 3

    def __init__(self, value=NVector(0, 0)):
        EffectValue.__init__(self)
        ## Effect value.
        self.value = MultiDimensional(value)


## @ingroup Lottie
class EffectValueCheckbox(EffectValue):
    _props = [
        LottieProp("value", "v", Value, False),
    ]
    ## %Effect type.
    type = 4

    def __init__(self, value=0):
        EffectValue.__init__(self)
        ## Effect value.
        self.value = Value(value)


## @ingroup Lottie
## @ingroup LottieCheck
## Lottie-web ignores these
class EffectValueIgnored(EffectValue):
    _props = [
        LottieProp("value", "v", float, False),
    ]
    ## %Effect type.
    type = 6

    def __init__(self, value=0):
        EffectValue.__init__(self)
        ## Effect value.
        self.value = value


## @ingroup Lottie
## @ingroup LottieCheck
class EffectValueDropDown(EffectValue):
    _props = [
        LottieProp("value", "v", Value, False),
    ]
    ## %Effect type.
    type = 7

    def __init__(self, value=0):
        EffectValue.__init__(self)
        ## Effect value.
        self.value = Value(value)


## @ingroup Lottie
## @ingroup LottieCheck
class EffectValueLayer(EffectValue):
    _props = [
        LottieProp("value", "v", Value, False),
    ]
    ## %Effect type.
    type = 10

    def __init__(self):
        EffectValue.__init__(self)
        ## Effect value.
        self.value = Value()


## @ingroup Lottie
class FillEffect(Effect):
    """!
    Replaces the whole layer with the given color
    @note Opacity is in [0, 1]
    """
    _effects = [
        ("00", EffectValuePoint),
        ("01", EffectValueDropDown),
        ("color", EffectValueColor),
        ("03", EffectValueDropDown),
        ("04", EffectValueSlider),
        ("05", EffectValueSlider),
        ("opacity", EffectValueSlider),
    ]
    ## %Effect type.
    type = 21


## @ingroup Lottie
class StrokeEffect(Effect):
    _effects = [
        ("00", EffectValueColor),
        ("01", EffectValueCheckbox),
        ("02", EffectValueCheckbox),
        ("color", EffectValueColor),
        ("04", EffectValueSlider),
        ("05", EffectValueSlider),
        ("06", EffectValueSlider),
        ("07", EffectValueSlider),
        ("08", EffectValueSlider),
        ("09", EffectValueDropDown),
        ("type", EffectValueDropDown),
    ]
    ## %Effect type.
    type = 22


## @ingroup Lottie
class TritoneEffect(Effect):
    """!
    Maps layers colors based on bright/mid/dark colors
    """
    _effects = [
        ("bright", EffectValueColor),
        ("mid", EffectValueColor),
        ("dark", EffectValueColor),
    ]
    ## %Effect type.
    type = 23


"""
## @ingroup Lottie
## @ingroup LottieCheck
class GroupEffect(Effect):
    _props = [
        LottieProp("enabled", "en", PseudoBool, False),
    ]

    def __init__(self):
        Effect.__init__(self)
        ## Enabled AE property value
        self.enabled = True
"""


## @ingroup Lottie
## @ingroup LottieCheck
class ProLevelsEffect(Effect):
    _effects = [
        ("00", EffectValueDropDown),
        ("01", EffectNoValue),
        ("02", EffectNoValue),
        ("comp_inblack", EffectValueSlider),
        ("comp_inwhite", EffectValueSlider),
        ("comp_gamma", EffectValueSlider),
        ("comp_outblack", EffectValueSlider),
        ("comp_outwhite", EffectValueSlider),
        ("08", EffectNoValue),
        ("09", EffectValueSlider),
        ("r_inblack", EffectValueSlider),
        ("r_inwhite", EffectValueSlider),
        ("r_gamma", EffectValueSlider),
        ("r_outblack", EffectValueSlider),
        ("r_outwhite", EffectValueSlider),
        ("15", EffectValueSlider),
        ("16", EffectValueSlider),
        ("g_inblack", EffectValueSlider),
        ("g_inwhite", EffectValueSlider),
        ("g_gamma", EffectValueSlider),
        ("g_outblack", EffectValueSlider),
        ("g_outwhite", EffectNoValue),
        ("22", EffectValueSlider),
        ("23", EffectValueSlider),
        ("b_inblack", EffectValueSlider),
        ("b_inwhite", EffectValueSlider),
        ("b_gamma", EffectValueSlider),
        ("b_outblack", EffectValueSlider),
        ("b_outwhite", EffectNoValue),
        ("29", EffectValueSlider),
        ("a_inblack", EffectValueSlider),
        ("a_inwhite", EffectValueSlider),
        ("a_gamma", EffectValueSlider),
        ("a_outblack", EffectValueSlider),
        ("a_outwhite", EffectNoValue),
    ]
    ## %Effect type.
    type = 24


## @ingroup Lottie
class TintEffect(Effect):
    """!
    Colorizes the layer
    @note Opacity is in [0, 100]
    """
    _effects = [
        ("color_black", EffectValueColor),
        ("color_white", EffectValueColor),
        ("opacity", EffectValueSlider),
    ]
    ## %Effect type.
    type = 20


## @ingroup Lottie
class DropShadowEffect(Effect):
    """!
    Adds a shadow to the layer
    @note Opacity is in [0, 255]
    """
    _effects = [
        ("color", EffectValueColor),
        ("opacity", EffectValueSlider),
        ("angle", EffectValueAngle),
        ("distance", EffectValueSlider),
        ("blur", EffectValueSlider),
    ]
    ## %Effect type.
    type = 25


## @ingroup Lottie
## @ingroup LottieCheck
class Matte3Effect(Effect):
    _effects = [
        ("index", EffectValueSlider),
    ]
    ## %Effect type.
    type = 28


## @ingroup Lottie
class GaussianBlurEffect(Effect):
    """!
    Gaussian blur
    """
    _effects = [
        ("sigma", EffectValueSlider),
        ("dimensions", EffectValueSlider),
        ("wrap", EffectValueCheckbox),
    ]
    ## %Effect type.
    type = 29


## @ingroup Lottie
class CustomEffect(Effect):
    """!
    Grouing properties in custom effects
    """
    _effects = []
    ## %Effect type.
    type = 5


#ingroup Lottie
class MeshWarpEffect(Effect):
    _effects = [
        ("Rows", EffectValueSlider),
        ("Columns", EffectValueSlider),
        ("Quality", EffectValueSlider),
        ("03", EffectNoValue),
    ]
    ## %Effect type.
    type = 31


#ingroup Lottie
class DisplacementMapEffect(Effect):
    _effects = [
        ("Displacement Map Layer", EffectValueLayer),
        ("Use For Horizontal Displacement", EffectValueDropDown),
        ("Max Horizontal Displacement", EffectValueSlider),
        ("Use For Vertical Displacement", EffectValueDropDown),
        ("Max Vertical Displacement", EffectValueSlider),
        ("Displacement Map Behavior", EffectValueDropDown),
        ("Edge Behavior", EffectValueDropDown),
        ("Expand Output", EffectValueDropDown),
    ]
    ## %Effect type.
    type = 27


#ingroup Lottie
class PaintOverTransparentEffect(Effect):
    _effects = [
        ("00", EffectValueSlider)
    ]
    ## %Effect type.
    type = 7


#ingroup Lottie
class PuppetEffect(Effect):
    _effects = [
        ("Puppet Engine", EffectValueDropDown),
        ("Mesh Rotation Refinement", EffectValueSlider),
        ("On Transparent", EffectValueDropDown),
        ("03", EffectNoValue),
    ]
    ## %Effect type.
    type = 34


#ingroup Lottie
class RadialWipeEffect(Effect):
    _effects = [
        ("Completion", EffectValueSlider),
        ("Start Angle", EffectValueAngle),
        ("Wipe Center", EffectValuePoint),
        ("Wipe", EffectValueSlider),
        ("Feather", EffectValueSlider),
    ]
    ## %Effect type.
    type = 26


#ingroup Lottie
class SpherizeEffect(Effect):
    _effects = [
        ("radius", EffectValueSlider),
        ("center", EffectValuePoint),
    ]
    ## %Effect type.
    effect_type = 33


#ingroup Lottie
class WavyEffect(Effect):
    _effects = [
        ("Radius", EffectValueSlider),
        ("Center", EffectValuePoint),
        ("Conversion type", EffectValueDropDown),
        ("Speed", EffectValueDropDown),
        ("Width", EffectValueSlider),
        ("Height", EffectValueSlider),
        ("Phase", EffectValueSlider),
    ]
    ## %Effect type.
    type = 32



import math
from functools import reduce
from .base import LottieObject, LottieProp, PseudoList, PseudoBool
from . import easing
from ..nvector import NVector
from .bezier import Bezier
from ..utils.color import Color


class KeyframeBezier:
    NEWTON_ITERATIONS = 4
    NEWTON_MIN_SLOPE = 0.001
    SUBDIVISION_PRECISION = 0.0000001
    SUBDIVISION_MAX_ITERATIONS = 10
    SPLINE_TABLE_SIZE = 11
    SAMPLE_STEP_SIZE = 1.0 / (SPLINE_TABLE_SIZE - 1.0)

    def __init__(self, h1, h2):
        self.h1 = h1
        self.h2 = h2
        self._sample_values = None

    @classmethod
    def from_keyframe(cls, keyframe):
        return cls(keyframe.out_value, keyframe.in_value)

    def bezier(self):
        bez = Bezier()
        bez.add_point(NVector(0, 0), outp=NVector(self.h1.x, self.h1.y))
        bez.add_point(NVector(1, 1), inp=NVector(self.h2.x-1, self.h2.y-1))
        return bez

    def _a(self, c1, c2):
        return 1 - 3 * c2 + 3 * c1

    def _b(self, c1, c2):
        return 3 * c2 - 6 * c1

    def _c(self, c1):
        return 3 * c1

    def _bezier_component(self, t, c1, c2):
        return ((self._a(c1, c2) * t + self._b(c1, c2)) * t + self._c(c1)) * t

    def point_at(self, t):
        return NVector(
            self._bezier_component(t, self.h1.x, self.h2.x),
            self._bezier_component(t, self.h1.y, self.h2.y)
        )

    def _slope_component(self, t, c1, c2):
        return 3 * self._a(c1, c2) * t * t + 2 * self._b(c1, c2) * t + self._c(c1)

    def slope_at(self, t):
        return NVector(
            self._slope_component(t, self.h1.x, self.h2.x),
            self._slope_component(t, self.h1.y, self.h2.y)
        )

    def _binary_subdivide(self, x, interval_start, interval_end):
        current_x = None
        t = None
        i = 0
        for i in range(self.SUBDIVISION_MAX_ITERATIONS):
            if current_x is not None and abs(current_x) < self.SUBDIVISION_PRECISION:
                break
            t = interval_start + (interval_end - interval_start) / 2.0
            current_x = self._bezier_component(t, self.h1.x, self.h2.x) - x
            if current_x > 0.0:
                interval_end = t
            else:
                interval_start = t
        return t

    def _newton_raphson(self, x, t_guess):
        for i in range(self.NEWTON_ITERATIONS):
            slope = self._slope_component(t_guess, self.h1.x, self.h2.x)
            if slope == 0:
                return t_guess
            current_x = self._bezier_component(t_guess, self.h1.x, self.h2.x) - x
            t_guess -= current_x / slope
        return t_guess

    def _get_sample_values(self):
        if self._sample_values is None:
            self._sample_values = [
                self._bezier_component(i * self.SAMPLE_STEP_SIZE, self.h1.x, self.h2.x)
                for i in range(self.SPLINE_TABLE_SIZE)
            ]
        return self._sample_values

    def t_for_x(self, x):
        sample_values = self._get_sample_values()
        interval_start = 0
        current_sample = 1
        last_sample = self.SPLINE_TABLE_SIZE - 1
        while current_sample != last_sample and sample_values[current_sample] <= x:
            interval_start += self.SAMPLE_STEP_SIZE
            current_sample += 1
        current_sample -= 1

        dist = (x - sample_values[current_sample]) / (sample_values[current_sample+1] - sample_values[current_sample])
        t_guess = interval_start + dist * self.SAMPLE_STEP_SIZE
        initial_slope = self._slope_component(t_guess, self.h1.x, self.h2.x)
        if initial_slope >= self.NEWTON_MIN_SLOPE:
            return self._newton_raphson(x, t_guess)
        if initial_slope == 0:
            return t_guess
        return self._binary_subdivide(x, interval_start, interval_start + self.SAMPLE_STEP_SIZE)

    def y_at_x(self, x):
        t = self.t_for_x(x)
        return self._bezier_component(t, self.h1.y, self.h2.y)


## @ingroup Lottie
class Keyframe(LottieObject):
    _props = [
        LottieProp("time", "t", float, False),
        LottieProp("in_value", "i", easing.KeyframeBezierHandle, False),
        LottieProp("out_value", "o", easing.KeyframeBezierHandle, False),
        LottieProp("hold", "h", PseudoBool),
    ]

    def __init__(self, time=0, easing_function=None):
        """!
        @param time             Start time of keyframe segment
        @param easing_function  Callable that performs the easing
        """
        ## Start time of keyframe segment.
        self.time = time
        ## Bezier curve easing in value.
        self.in_value = None
        ## Bezier curve easing out value.
        self.out_value = None
        ## Jump to the end value
        self.hold = None

        if easing_function:
            easing_function(self)

    @property
    def jump(self):
        return self.hold

    @jump.setter
    def jump(self, v):
        self.hold = v

    def bezier(self):
        if self.hold:
            bez = Bezier()
            bez.add_point(NVector(0, 0))
            bez.add_point(NVector(1, 0))
            bez.add_point(NVector(1, 1))
            return bez
        else:
            return KeyframeBezier.from_keyframe(self).bezier()

    def lerp_factor(self, ratio):
        return KeyframeBezier.from_keyframe(self).y_at_x(ratio)

    def __str__(self):
        return "%s %s" % (self.time, self.start)


## @ingroup Lottie
class OffsetKeyframe(Keyframe):
    """!
    Keyframe for MultiDimensional values

    @par Bezier easing
    @parblock
    Imagine a quadratic bezier, with starting point at (0, 0) and end point at (1, 1).

    @p out_value and @p in_value are the other two handles for a quadratic bezier,
    expressed as absolute values in this 0-1 space.

    See also https://cubic-bezier.com/
    @endparblock
    """
    _props = [
        LottieProp("start", "s", NVector, False),
        LottieProp("end", "e", NVector, False),
    ]

    def __init__(self, time=0, start=None, end=None, easing_function=None, in_tan=None, out_tan=None):
        Keyframe.__init__(self, time, easing_function)
        ## Start value of keyframe segment.
        self.start = start
        ## End value of keyframe segment.
        self.end = end
        ## In Spatial Tangent. Only for spatial properties. (for bezier smoothing on position)
        self.in_tan = in_tan
        ## Out Spatial Tangent. Only for spatial properties. (for bezier smoothing on position)
        self.out_tan = out_tan

    def interpolated_value(self, ratio, next_start=None):
        end = next_start if self.end is None else self.end
        if end is None:
            return self.start
        if not self.in_value or not self.out_value:
            return self.start
        if ratio == 1:
            return end
        if ratio == 0:
            return self.start
        if self.in_tan and self.out_tan:
            bezier = Bezier()
            bezier.add_point(self.start, NVector(0, 0), self.out_tan)
            bezier.add_point(end, self.in_tan, NVector(0, 0))
            return bezier.point_at(ratio)

        lerpv = self.lerp_factor(ratio)
        return self.start.lerp(end, lerpv)

    def interpolated_tangent_angle(self, ratio, next_start=None):
        end = next_start if self.end is None else self.end
        if end is None or not self.in_tan or not self.out_tan:
            return 0

        bezier = Bezier()
        bezier.add_point(self.start, NVector(0, 0), self.out_tan)
        bezier.add_point(end, self.in_tan, NVector(0, 0))
        return bezier.tangent_angle_at(ratio)

    def __repr__(self):
        return "<%s.%s %s %s%s>" % (
            type(self).__module__,
            type(self).__name__,
            self.time,
            self.start,
            (" -> %s" % self.end) if self.end is not None else ""
        )


class AnimatableMixin:
    keyframe_type = Keyframe

    def __init__(self, value=None):
        ## Non-animated value
        self.value = value
        ## Property index
        self.property_index = None
        ## Whether it's animated
        self.animated = False
        ## Keyframe list
        self.keyframes = None
        ## Expression
        self.expression = None

    def clear_animation(self, value):
        """!
        Sets a fixed value, removing animated keyframes
        """
        self.value = value
        self.animated = False
        self.keyframes = None

    def add_keyframe(self, time, value, interp=easing.Linear(), *args, **kwargs):
        """!
        @param time     The time this keyframe appears in
        @param value    The value the property should have at @p time
        @param interp   The easing callable used to update the tangents of the previous keyframe
        @param args     Extra arguments to pass the keyframe constructor
        @param kwargs   Extra arguments to pass the keyframe constructor
        @note Always call add_keyframe with increasing @p time value
        """
        if not self.animated:
            self.value = None
            self.keyframes = []
            self.animated = True
        else:
            if self.keyframes[-1].time == time:
                if value != self.keyframes[-1].start:
                    self.keyframes[-1].start = value
                return
            else:
                self.keyframes[-1].end = value.clone()

        self.keyframes.append(self.keyframe_type(
            time,
            value,
            None,
            interp,
            *args,
            **kwargs
        ))

    def get_value(self, time=0):
        """!
        @brief Returns the value of the property at the given frame/time
        """
        if not self.animated:
            return self.value

        if not self.keyframes:
            return None

        return self._get_value_helper(time)[0]

    def _get_value_helper(self, time):
        val = self.keyframes[0].start
        for i in range(len(self.keyframes)):
            k = self.keyframes[i]
            if time - k.time <= 0:
                if k.start is not None:
                    val = k.start

                kp = self.keyframes[i-1] if i > 0 else None
                if kp:
                    t = (time - kp.time) / (k.time - kp.time)
                    end = kp.end
                    if end is None:
                        end = val
                    if end is not None:
                        val = kp.interpolated_value(t, end)
                    return val, end, kp, t
                return val, None, None, None
            if k.end is not None:
                val = k.end
        return val, None, None, None

    def to_dict(self):
        d = super().to_dict()
        if self.animated:
            if "k" not in d:
                return d
            last = d["k"][-1]
            last.pop("i", None)
            last.pop("o", None)
        return d

    def __repr__(self):
        if self.keyframes and len(self.keyframes) > 1:
            val = "%s -> %s" % (self.keyframes[0].start, self.keyframes[-2].end)
        else:
            val = self.value
        return "<%s.%s %s>" % (type(self).__module__, type(self).__name__, val)

    def __str__(self):
        if self.animated:
            return "animated"
        return str(self.value)

    @classmethod
    def merge_keyframes(cls, items, conversion):
        """
        @todo Remove similar functionality from SVG/sif parsers
        """
        keyframes = []
        for animatable in items:
            if animatable.animated:
                keyframes.extend(animatable.keyframes)

        # TODO properly interpolate tangents
        new_kframes = []
        for keyframe in sorted(keyframes, key=lambda kf: kf.time):
            if new_kframes and new_kframes[-1].time == keyframe.time:
                continue
            kfcopy = keyframe.clone()
            kfcopy.start = conversion(*(i.get_value(keyframe.time) for i in items))
            new_kframes.append(kfcopy)

        for i in range(0, len(new_kframes) - 1):
            new_kframes[i].end = new_kframes[i+1].start

        return new_kframes

    @classmethod
    def load(cls, lottiedict):
        obj = super().load(lottiedict)
        if "a" not in lottiedict:
            obj.animated = prop_animated(lottiedict)
        return obj

    def average(self, start, end, value_map=lambda v: v):
        if not self.animated or len(self.keyframes) == 0:
            return value_map(self.value)
        elif len(self.keyframes) == 1:
            return value_map(self.keyframes[0].start)

        total_weight = 0
        value = value_map(self.keyframes[0].start) * 0
        kf_index = 0

        if self.keyframes[0].time > start:
            weight = self.keyframes[0].time - start
            value += value_map(self.keyframes[0].start) * weight
            total_weight += weight
        elif self.keyframes[0].time < start:
            avg = (self.keyframes[1].time + self.keyframes[0].time) / 2
            kf_index = 1
            if self.keyframes[0].hold:
                weight = self.keyframes[1].time - start
                value += value_map(self.keyframes[0].start) * weight
                total_weight += weight
            elif start < avg:
                weight = avg - start
                value += value_map(self.keyframes[0].start) * weight
                total_weight += weight

                half = (self.keyframes[1].time - self.keyframes[0].time) / 2
                value += value_map(self.keyframes[1].start) * half
                total_weight += half
            else:
                weight = self.keyframes[1].time - start
                value += value_map(self.keyframes[1].start) * weight
                total_weight += weight

        for i in range(kf_index, len(self.keyframes) - 1):
            kf = self.keyframes[i]
            kfn = self.keyframes[i + 1]
            if kfn.time > end:
                break

            delta = kfn.time - kf.time
            total_weight += delta
            if kf.hold:
                value += value_map(kf.start) * delta
            else:
                value += value_map(kf.start) * (delta / 2)
                value += value_map(kfn.start) * (delta / 2)
        else:
            kf = self.keyframes[-1]
            kfn = self.keyframes[-1]

        if kfn.time < end:
            weight = end - kfn.time
            value += value_map(kfn.start) * weight
            total_weight += weight
        elif kf.hold:
            weight = end - kf.time
            value += value_map(kf.start) * weight
            total_weight += weight
        elif kfn.time > end:
            avg = (kf.time + kfn.time) / 2
            if end < avg:
                weight = end - kf.time
                value += value_map(kf.start) * weight
                total_weight += weight
            else:
                half = (kfn.time - kf.time) / 2
                value += value_map(kf.start) * half
                total_weight += half

                weight = end - avg
                value += value_map(kfn.start) * weight
                total_weight += weight

        if total_weight == 0:
            return value
        return value / total_weight


def prop_animated(l):
    if "a" in l:
        return l["a"]
    if "k" not in l:
        return False
    if isinstance(l["k"], list) and l["k"] and isinstance(l["k"][0], dict):
        return True
    return False


def prop_not_animated(l):
    return not prop_animated(l)


## @ingroup Lottie
class MultiDimensional(AnimatableMixin, LottieObject):
    """!
    An animatable property that holds a NVector
    """
    keyframe_type = OffsetKeyframe
    _props = [
        LottieProp("value", "k", NVector, False, prop_not_animated),
        LottieProp("property_index", "ix", int, False),
        LottieProp("animated", "a", PseudoBool, False),
        LottieProp("keyframes", "k", OffsetKeyframe, True, prop_animated),
        LottieProp("expression", "x", str, False),
    ]

    def get_tangent_angle(self, time=0):
        """!
        @brief Returns the value tangent angle of the property at the given frame/time
        """
        if not self.keyframes or len(self.keyframes) < 2:
            return 0

        val, end, kp, t = self._get_value_helper(time)
        if kp:
            return kp.interpolated_tangent_angle(t, end)

        if self.keyframes[0].time >= time:
            end = self.keyframes[0].end if self.keyframes[0].end is not None else self.keyframes[1].start
            return self.keyframes[0].interpolated_tangent_angle(0, end)

        return 0


## @ingroup Lottie
class PositionKeyframe(OffsetKeyframe):
    """!
    Keyframe for Positional values
    """
    _props = [
        LottieProp("in_tan", "ti", NVector, False),
        LottieProp("out_tan", "to", NVector, False),
    ]


class PositionValue(MultiDimensional):
    keyframe_type = PositionKeyframe
    _props = [
        LottieProp("value", "k", NVector, False, prop_not_animated),
        LottieProp("property_index", "ix", int, False),
        LottieProp("animated", "a", PseudoBool, False),
        LottieProp("keyframes", "k", OffsetKeyframe, True, prop_animated),
    ]

    @classmethod
    def load(cls, lottiedict):
        obj = super().load(lottiedict)
        if lottiedict.get("s", False):
            cls._load_split(lottiedict, obj)

        return obj

    @classmethod
    def _load_split(cls, lottiedict, obj):
        components = [
            Value.load(lottiedict.get("x", {})),
            Value.load(lottiedict.get("y", {})),
        ]
        if "z" in lottiedict:
            components.append(Value.load(lottiedict.get("z", {})))

        has_anim = any(x for x in components if x.animated)
        if not has_anim:
            obj.value = NVector(*(a.value for a in components))
            obj.animated = False
            obj.keyframes = None
            return

        obj.animated = True
        obj.value = None
        obj.keyframes = cls.merge_keyframes(components, NVector)


class ColorValue(AnimatableMixin, LottieObject):
    """!
    An animatable property that holds a Color
    """
    keyframe_type = OffsetKeyframe
    _props = [
        LottieProp("value", "k", Color, False, prop_not_animated),
        LottieProp("property_index", "ix", int, False),
        LottieProp("animated", "a", PseudoBool, False),
        LottieProp("keyframes", "k", OffsetKeyframe, True, prop_animated),
        LottieProp("expression", "x", str, False),
    ]


## @ingroup Lottie
class GradientColors(LottieObject):
    """!
    Represents colors and offsets in a gradient

    Colors are represented as a flat list interleaving offsets and color components in weird ways
    There are two possible layouts:

    Without alpha, the colors are a sequence of offset, r, g, b

    With alpha, same as above but at the end of the list there is a sequence of offset, alpha

    Examples:

    For the gradient [0, red], [0.5, yellow], [1, green]
    The list would be [0, 1, 0, 0, 0.5, 1, 1, 0, 1, 0, 1, 0]

    For the gradient [0, red at 80% opacity], [0.5, yellow at 70% opacity], [1, green at 60% opacity]
    The list would be [0, 1, 0, 0, 0.5, 1, 1, 0, 1, 0, 1, 0, 0, 0.8, 0.5, 0.7, 1, 0.6]
    """
    _props = [
        LottieProp("colors", "k", MultiDimensional),
        LottieProp("count", "p", int),
    ]

    def __init__(self, stops=[]):
        ## Animatable colors, as a vector containing [offset, r, g, b] values as a flat array
        self.colors = MultiDimensional(NVector())
        ## Number of colors
        self.count = 0
        if stops:
            self.set_stops(stops)

    @staticmethod
    def color_to_stops(self, colors):
        """
        Converts a list of colors (Color) to tuples (offset, color)
        """
        return [
            (i / (len(colors)-1), color)
            for i, color in enumerate(colors)
        ]

    def set_stops(self, stops, keyframe=None):
        """!
        @param stops iterable of (offset, Color) tuples
        @param keyframe keyframe index (or None if not animated)
        """
        flat = self._flatten_stops(stops)
        if self.colors.animated and keyframe is not None:
            if keyframe > 1:
                self.colors.keyframes[keyframe-1].end = flat
            self.colors.keyframes[keyframe].start = flat
        else:
            self.colors.clear_animation(flat)
        self.count = len(stops)

    def _flatten_stops(self, stops):
        flattened_colors = NVector(*reduce(
            lambda a, b: a + b,
            (
                [off] + color.components[:3]
                for off, color in stops
            )
        ))

        if any(len(c) > 3 for o, c in stops):
            flattened_colors.components += reduce(
                lambda a, b: a + b,
                (
                    [off] + [self._get_alpha(color)]
                    for off, color in stops
                )
            )
        return flattened_colors

    def _get_alpha(self, color):
        if len(color) > 3:
            return color[3]
        return 1

    def _add_to_flattened(self, offset, color, flattened):
        flat = [offset] + list(color[:3])
        rgb_size = 4 * self.count

        if len(flattened) == rgb_size:
            # No alpha
            flattened.extend(flat)
            if self.count == 0 and len(color) > 3:
                flattened.append(offset)
                flattened.append(color[3])
        else:
            flattened[rgb_size:rgb_size] = flat
            flattened.append(offset)
            flattened.append(self._get_alpha(color))

    def add_color(self, offset, color, keyframe=None):
        if self.colors.animated:
            if keyframe is None:
                for kf in self.colors.keyframes:
                    if kf.start:
                        self._add_to_flattened(offset, color, kf.start.components)
                    if kf.end:
                        self._add_to_flattened(offset, color, kf.end.components)
            else:
                if keyframe > 1:
                    self._add_to_flattened(offset, color, self.colors.keyframes[keyframe-1].end.components)
                self._add_to_flattened(offset, color, self.colors.keyframes[keyframe].start.components)
        else:
            self._add_to_flattened(offset, color, self.colors.value.components)
        self.count += 1

    def add_keyframe(self, time, stops, ease=easing.Linear()):
        """!
        @param time   Frame time
        @param stops  Iterable of (offset, Color) tuples
        @param ease   Easing function
        """
        self.colors.add_keyframe(time, self._flatten_stops(stops), ease)

    def get_stops(self, keyframe=0):
        if keyframe is not None:
            colors = self.colors.keyframes[keyframe].start
        else:
            colors = self.colors.value
        return self._stops_from_flat(colors)

    def _stops_from_flat(self, colors):
        if len(colors) == 4 * self.count:
            for i in range(self.count):
                off = i * 4
                yield colors[off], Color(*colors[off+1:off+4])
        else:
            for i in range(self.count):
                off = i * 4
                aoff = self.count * 4 + i * 2 + 1
                yield colors[off], Color(colors[off+1], colors[off+2], colors[off+3], colors[aoff])

    def stops_at(self, time):
        return self._stops_from_flat(self.colors.get_value(time))


## @ingroup Lottie
class Value(AnimatableMixin, LottieObject):
    """!
    An animatable property that holds a float
    """
    keyframe_type = OffsetKeyframe
    _props = [
        LottieProp("value", "k", float, False, prop_not_animated),
        LottieProp("property_index", "ix", int, False),
        LottieProp("animated", "a", PseudoBool, False),
        LottieProp("keyframes", "k", keyframe_type, True, prop_animated),
        LottieProp("expression", "x", str, False),
    ]

    def __init__(self, value=0):
        super().__init__(value)

    def add_keyframe(self, time, value, ease=easing.Linear()):
        super().add_keyframe(time, NVector(value), ease)

    def get_value(self, time=0):
        v = super().get_value(time)
        if self.animated and self.keyframes:
            return v[0]
        return v


## @ingroup Lottie
class ShapePropKeyframe(Keyframe):
    """!
    Keyframe holding Bezier objects
    """
    _props = [
        LottieProp("start", "s", Bezier, PseudoList),
        LottieProp("end", "e", Bezier, PseudoList),
    ]

    def __init__(self, time=0, start=None, end=None, easing_function=None):
        Keyframe.__init__(self, time, easing_function)
        ## Start value of keyframe segment.
        self.start = start
        ## End value of keyframe segment.
        self.end = end

    def interpolated_value(self, ratio, next_start=None):
        end = next_start if self.end is None else self.end
        if end is None:
            return self.start
        if not self.in_value or not self.out_value:
            return self.start
        if ratio == 1:
            return end
        if ratio == 0 or len(self.start.vertices) != len(end.vertices):
            return self.start

        lerpv = self.lerp_factor(ratio)
        bez = Bezier()
        bez.closed = self.start.closed
        for i in range(len(self.start.vertices)):
            bez.vertices.append(self.start.vertices[i].lerp(end.vertices[i], lerpv))
            bez.in_tangents.append(self.start.in_tangents[i].lerp(end.in_tangents[i], lerpv))
            bez.out_tangents.append(self.start.out_tangents[i].lerp(end.out_tangents[i], lerpv))
        return bez


## @ingroup Lottie
class ShapeProperty(AnimatableMixin, LottieObject):
    """!
    An animatable property that holds a Bezier
    """
    keyframe_type = ShapePropKeyframe
    _props = [
        LottieProp("value", "k", Bezier, False, prop_not_animated),
        LottieProp("expression", "x", str, False),
        LottieProp("property_index", "ix", float, False),
        LottieProp("animated", "a", PseudoBool, False),
        LottieProp("keyframes", "k", keyframe_type, True, prop_animated),
    ]

    def __init__(self, bezier=None):
        super().__init__(bezier or Bezier())


#ingroup Lottie
class SplitVector(LottieObject):
    """!
    An animatable property that is split into individually anaimated components
    """
    _props = [
        LottieProp("split", "s", bool, False),
        LottieProp("x", "x", Value, False),
        LottieProp("y", "y", Value, False),
        LottieProp("z", "z", Value, False),
    ]

    @property
    def split(self):
        return True

    def __init__(self, x=0, y=0):
        super().__init__()

        self.x = Value(x)
        self.y = Value(y)
        self.z = None




from .base import exporter
from ..parsers.sif.builder import to_sif
from ..utils.file import open_file


@exporter("Synfig", ["sif"], [], {"pretty"})
def export_sif(animation, file, pretty=True):
    with open_file(file) as fp:
        dom = to_sif(animation).to_xml()
        dom.writexml(fp, "", "  " if pretty else "", "\n" if pretty else "")



from .base import LottieEnum


## @ingroup Lottie
class TextBased(LottieEnum):
    Characters = 1
    CharacterExcludingSpaces = 2
    Words = 3
    Lines = 4

    @classmethod
    def default(cls):
        return cls.Characters


## @ingroup Lottie
class TextShape(LottieEnum):
    Square = 1
    RampUp = 2
    RampDown = 3
    Triangle = 4
    Round = 5
    Smooth = 6

    @classmethod
    def default(cls):
        return cls.Square



import math

from ..nvector import NVector, PolarVector
from ..objects.bezier import BezierPoint, Bezier


## @todo Just output a Bezier object
class Ellipse:
    def __init__(self, center, radii, xrot):
        """
        @param center      2D vector, center of the ellipse
        @param radii       2D vector, x/y radius of the ellipse
        @param xrot        Angle between the main axis of the ellipse and the x axis (in radians)
        """
        self.center = center
        self.radii = radii
        self.xrot = xrot

    def point(self, t):
        return NVector(
            self.center[0]
            + self.radii[0] * math.cos(self.xrot) * math.cos(t)
            - self.radii[1] * math.sin(self.xrot) * math.sin(t),

            self.center[1]
            + self.radii[0] * math.sin(self.xrot) * math.cos(t)
            + self.radii[1] * math.cos(self.xrot) * math.sin(t)
        )

    def derivative(self, t):
        return NVector(
            - self.radii[0] * math.cos(self.xrot) * math.sin(t)
            - self.radii[1] * math.sin(self.xrot) * math.cos(t),

            - self.radii[0] * math.sin(self.xrot) * math.sin(t)
            + self.radii[1] * math.cos(self.xrot) * math.cos(t)
        )

    def to_bezier_points(self, anglestart, angle_delta, step=math.pi / 2):
        points = []
        angle1 = anglestart
        angle_left = abs(angle_delta)
        sign = -1 if anglestart+angle_delta < angle1 else 1
        tolerance = math.pi / 100
        if angle_left % step > tolerance:
            step = angle_left / max(1, round(angle_left / step))

        # We need to fix the first handle
        firststep = min(angle_left, step) * sign
        alpha = self._alpha(firststep)
        q1 = self.derivative(angle1) * alpha
        points.append(BezierPoint(self.point(angle1), NVector(0, 0), q1))

        # Then we iterate until the angle has been completed
        half_step = step / 2
        while True:
            if angle_left < half_step:
                break

            lstep = min(angle_left, step)
            step_sign = lstep * sign
            angle2 = angle1 + step_sign
            angle_left -= abs(lstep)

            alpha = self._alpha(step_sign)
            p2 = self.point(angle2)
            q2 = self.derivative(angle2) * alpha

            points.append(BezierPoint(p2, -q2, q2))
            angle1 = angle2

        return points

    def to_bezier(self, angle_start, angle_delta, step=math.pi / 2):
        bezier = Bezier()
        points = self.to_bezier_points(angle_start, angle_delta, step)

        if angle_delta == math.pi * 2:
            points.pop(0)
            bezier.close()

        for point in points:
            bezier.add_point(point.vertex, point.in_tangent, point.out_tangent)
        return bezier

    def _alpha(self, step):
        return math.sin(step) * (math.sqrt(4+3*math.tan(step/2)**2) - 1) / 3

    @classmethod
    def from_svg_arc(cls, start, rx, ry, xrot, large, sweep, dest):
        rx = abs(rx)
        ry = abs(ry)

        x1 = start[0]
        y1 = start[1]
        x2 = dest[0]
        y2 = dest[1]
        phi = math.pi * xrot / 180

        x1p, y1p = _matrix_mul(phi, (start-dest)/2, -1)

        cr = x1p ** 2 / rx**2 + y1p**2 / ry**2
        if cr > 1:
            s = math.sqrt(cr)
            rx *= s
            ry *= s

        dq = rx**2 * y1p**2 + ry**2 * x1p**2
        pq = (rx**2 * ry**2 - dq) / dq
        cpm = math.sqrt(max(0, pq))
        if large == sweep:
            cpm = -cpm
        cp = NVector(cpm * rx * y1p / ry, -cpm * ry * x1p / rx)
        c = _matrix_mul(phi, cp) + NVector((x1+x2)/2, (y1+y2)/2)
        theta1 = _angle(NVector(1, 0), NVector((x1p - cp[0]) / rx, (y1p - cp[1]) / ry))
        deltatheta = _angle(
            NVector((x1p - cp[0]) / rx, (y1p - cp[1]) / ry),
            NVector((-x1p - cp[0]) / rx, (-y1p - cp[1]) / ry)
        ) % (2*math.pi)

        if not sweep and deltatheta > 0:
            deltatheta -= 2*math.pi
        elif sweep and deltatheta < 0:
            deltatheta += 2*math.pi

        return cls(c, NVector(rx, ry), phi), theta1, deltatheta


def _matrix_mul(phi, p, sin_mul=1):
    c = math.cos(phi)
    s = math.sin(phi) * sin_mul

    xr = c * p.x - s * p.y
    yr = s * p.x + c * p.y
    return NVector(xr, yr)


def _angle(u, v):
    arg = math.acos(max(-1, min(1, u.dot(v) / (u.length * v.length))))
    if u[0] * v[1] - u[1] * v[0] < 0:
        return -arg
    return arg



from .base import importer
from ..parsers.baseporter import ExtraOption
from ..parsers.pixel import (
    pixel_to_animation_paths, pixel_to_animation,
    raster_to_embedded_assets, raster_to_linked_assets
)
from ..parsers.svg.importer import parse_color

try:
    from ..parsers.raster import raster_to_animation, TraceOptions
    raster = True
except ImportError:
    raster = False


@importer("Raster image", ["bmp", "png", "gif", "webp", "tiff"], [
    ExtraOption("n_colors", type=int, default=1, help="Number of colors to quantize"),
    ExtraOption("palette", type=parse_color, default=[], nargs="+", help="Custom palette"),
    ExtraOption(
        "mode",
        default="trace" if raster else "embed",
        choices=["external", "embed", "pixel", "polygon"] + (["trace"] if raster else []),
        help="Vectorization mode:\n" +
        " * external : load images as linked assets\n" +
        " * embed    : load images as embedded assets\n" +
        " * pixel    : Vectorize the image into rectangles\n" +
        " * polygon  : Vectorize the image into polygonal shapes\n" +
        "              Looks the same as pixel, but a single shape per color\n" +
        " * trace    : (if available) Use potrace to vectorize\n"
    ),
    ExtraOption("frame_delay", type=int, default=4, help="Number of frames to skip between images"),
    ExtraOption("framerate", type=int, default=60, help="Frames per second"),
    ExtraOption("frame_files", nargs="+", default=[], help="Additional frames to import"),
    ExtraOption(
        "color_mode",
        default="nearest",
        choices=["nearest", "exact"],
        help="How to quantize colors.\n" +
             " * nearest    will map each color to the most similar in the palette\n" +
             " * exact      will only match exact colors"
    ),
    ExtraOption(
        "embed_format",
        default=None,
        help="Format to store images internally when using `embed` mode"
    ),
    ExtraOption(
        "stroke",
        default=1,
        type=int,
        help="Stroke width for output shapes"
    ),
])
def import_raster(filenames, n_colors, palette, mode, frame_delay=1,
                  framerate=60, frame_files=[], color_mode="nearest",
                  embed_format=None, stroke=1):
    if not isinstance(filenames, list):
        filenames = [filenames]
    filenames = filenames + frame_files

    if mode == "embed":
        return raster_to_embedded_assets(filenames, frame_delay, framerate, embed_format)
    elif mode == "external":
        return raster_to_linked_assets(filenames, frame_delay, framerate)
    elif mode == "trace":
        from ..parsers.raster import QuanzationMode
        options = TraceOptions()
        options.color_mode = QuanzationMode.Nearest if color_mode == "nearest" else QuanzationMode.Exact
        options.stroke_width = stroke
        return raster_to_animation(
            filenames, n_colors, frame_delay,
            framerate=framerate,
            palette=palette,
            trace_options=options
        )
    elif mode == "polygon":
        return pixel_to_animation_paths(filenames, frame_delay, framerate)
    else:
        return pixel_to_animation(filenames, frame_delay, framerate)



import math
from .base import LottieObject, LottieProp, PseudoList, PseudoBool


## @ingroup Lottie
class KeyframeBezierHandle(LottieObject):
    """!
    Bezier handle for keyframe interpolation
    """
    _props = [
        LottieProp("x", "x", list=PseudoList),
        LottieProp("y", "y", list=PseudoList),
    ]

    def __init__(self, x=0, y=0):
        ## x position of the handle.
        ## This represents the change in time of the keyframe
        self.x = x
        ## y position of the handle.
        ## This represents the change in value of the keyframe
        self.y = y


class Linear:
    """!
    Linear easing, the value will change from start to end in a straight line
    """
    def __call__(self, keyframe):
        keyframe.out_value = KeyframeBezierHandle(
            0,
            0
        )
        keyframe.in_value = KeyframeBezierHandle(
            1,
            1
        )


class EaseIn:
    """!
    The value lingers near the start before accelerating towards the end
    """
    def __init__(self, delay=1/3):
        self.delay = delay

    def __call__(self, keyframe):
        keyframe.out_value = KeyframeBezierHandle(
            self.delay,
            0
        )
        keyframe.in_value = KeyframeBezierHandle(
            1,
            1
        )


class EaseOut:
    """!
    The value starts fast before decelerating towards the end
    """
    def __init__(self, delay=1/3):
        self.delay = delay

    def __call__(self, keyframe):
        keyframe.out_value = KeyframeBezierHandle(
            0,
            0
        )
        keyframe.in_value = KeyframeBezierHandle(
            1-self.delay,
            1
        )


class Jump:
    """!
    Jumps to the end value at the end of the keyframe
    """
    def __call__(self, keyframe):
        keyframe.jump = True


class Sigmoid:
    """!
    Combines the effects of EaseIn and EaseOut
    """
    def __init__(self, delay=1/3):
        self.delay = delay

    def __call__(self, keyframe):
        keyframe.out_value = KeyframeBezierHandle(
            self.delay,
            0
        )
        keyframe.in_value = KeyframeBezierHandle(
            1 - self.delay,
            1
        )


class Split:
    """
    Uses different easing methods for in/out
    """

    def __init__(self, out_ease, in_ease):
        self.out_ease = out_ease
        self.in_ease = in_ease

    def __call__(self, keyframe):
        self.out_ease(keyframe)
        t = keyframe.out_value
        self.in_ease(keyframe)
        keyframe.out_value = t



from .base import LottieObject, LottieProp, LottieEnum
from .properties import Value, MultiDimensional, ColorValue, Color
from ..nvector import NVector
from .helpers import Transform
from .shapes import ShapeElement


#ingroup Lottie
class MaskedPath(LottieObject):
    """!
    Uses the path described by a layer mask to put the text on said path.
    """
    _props = [
        LottieProp("mask", "m", int, False),
        LottieProp("first_margin", "f", Value, False),
        LottieProp("last_margin", "l", Value, False),
        LottieProp("reverse_path", "r", Value, False),
        LottieProp("force_alignment", "a", Value, False),
        LottieProp("perpendicular_to_path", "p", Value, False),
    ]

    def __init__(self):
        super().__init__()

        ## Index of the mask to use
        self.mask = None
        self.first_margin = None
        self.last_margin = None
        self.reverse_path = None
        self.force_alignment = None
        self.perpendicular_to_path = None


## @ingroup Lottie
## @ingroup LottieCheck
class TextAnimatorDataProperty(Transform):
    _props = [
        LottieProp("rotate_x", "rx", Value),
        LottieProp("rotate_y", "ry", Value),

        LottieProp("stroke_width", "sw", Value),
        LottieProp("stroke_color", "sc", ColorValue),
        LottieProp("stroke_hue", "sh", Value, False),
        LottieProp("stroke_saturation", "ss", Value, False),
        LottieProp("stroke_brightness", "sb", Value, False),
        LottieProp("stroke_opacity", "so", Value, False),

        LottieProp("fill_color", "fc", ColorValue),
        LottieProp("fill_hue", "fh", Value),
        LottieProp("fill_saturation", "fs", Value),
        LottieProp("fill_brightness", "fb", Value),
        LottieProp("fill_opacity", "fo", Value, False),

        LottieProp("tracking", "t", Value),
        LottieProp("scale", "s", MultiDimensional),
        LottieProp("blur", "bl", Value, False),
        LottieProp("line_spacing", "ls", Value, False),

    ]

    def __init__(self):
        super().__init__()
        ## Angle?
        self.rx = Value()
        ## Angle?
        self.ry = Value()

        ## Stroke width
        self.stroke_width = Value()
        ## Stroke color
        self.stroke_color = ColorValue()
        self.stroke_hue = None
        self.stroke_saturation = None
        self.stroke_brightness = None
        self.stroke_opacity = None

        ## Fill color
        self.fill_color = ColorValue()
        ## Hue
        self.fill_hue = Value()
        ## Saturation 0-100
        self.fill_saturation = Value()
        ## Brightness 0-100
        self.fill_brightness = Value()
        self.fill_opacity = None

        ## Tracking
        self.tracking = Value()
        self.blur = None
        self.line_spacing = None


## @ingroup Lottie
class TextGrouping(LottieEnum):
    Characters = 1
    Word = 2
    Line = 3
    All = 4

    @classmethod
    def default(cls):
        return cls.Characters


## @ingroup Lottie
## @ingroup LottieCheck
class TextMoreOptions(LottieObject):
    _props = [
        LottieProp("alignment", "a", MultiDimensional),
        LottieProp("grouping", "g", TextGrouping),
    ]

    def __init__(self):
        self.alignment = MultiDimensional(NVector(0, 0))
        self.grouping = TextGrouping.default()


## @ingroup Lottie
class TextJustify(LottieEnum):
    Left = 0
    Right = 1
    Center = 2
    JustifyWithLastLineLeft = 3
    JustifyWithLastLineRight = 4
    JustifyWithLastLineCenter = 5
    JustifyWithLastLineFull = 6


#ingroup Lottie
class TextCaps(LottieEnum):
    Regular = 0
    AllCaps = 1
    SmallCaps = 2


## @ingroup Lottie
class TextDocument(LottieObject):
    """!
    @see http://docs.aenhancers.com/other/textdocument/

    Note that for multi-line text, lines are separated by \\r
    """
    _props = [
        LottieProp("font_family", "f", str),
        LottieProp("fill_color", "fc", Color),

        LottieProp("stroke_color", "sc", Color, False),
        LottieProp("stroke_width", "sw", float, False),
        LottieProp("stroke_over_fill", "of", bool, False),

        LottieProp("font_size", "s", float),
        LottieProp("line_height", "lh", float),
        LottieProp("wrap_size", "sz", NVector),
        LottieProp("wrap_position", "ps", NVector, False),

        LottieProp("text", "t", str),
        LottieProp("justify", "j", TextJustify),
        LottieProp("text_caps", "ca", TextCaps, False),
        LottieProp("tracking", "tr", float, False),
        LottieProp("baseline_shift", "ls", float, False),
    ]

    def __init__(self, text="", font_size=10, color=None, font_family=""):
        self.font_family = font_family

        ## Text color
        self.fill_color = color or Color(0, 0, 0)

        self.stroke_color = None
        self.stroke_width = 0
        ## Render stroke above the fill
        self.stroke_over_fill = None

        ## Line height when wrapping
        self.line_height = None
        ## Text alignment
        self.justify = TextJustify.Left
        ## Size of the box containing the text
        self.wrap_size = None
        ## Position of the box containing the text
        self.wrap_position = None
        ## Text
        self.text = text
        ## Font Size
        self.font_size = font_size
        ## Text caps
        self.text_caps = None
        ## Text Tracking
        self.tracking = None
        self.baseline_shift = None


## @ingroup Lottie
class TextDataKeyframe(LottieObject):
    _props = [
        LottieProp("start", "s", TextDocument),
        LottieProp("time", "t", float),
    ]

    def __init__(self, time=0, start=None):
        ## Start value of keyframe segment.
        self.start = start
        ## Start time of keyframe segment.
        self.time = time


## @ingroup Lottie
class TextData(LottieObject):
    _props = [
        LottieProp("keyframes", "k", TextDataKeyframe, True),
    ]

    def __init__(self):
        self.keyframes = []

    def get_value(self, time):
        for kf in self.keyframes:
            if kf.time >= time:
                return kf.start
        return None


## @ingroup Lottie
class TextAnimatorData(LottieObject):
    _props = [
        LottieProp("properties", "a", TextAnimatorDataProperty, True),
        LottieProp("data", "d", TextData, False),
        LottieProp("more_options", "m", TextMoreOptions, False),
        LottieProp("masked_path", "p", MaskedPath),
    ]

    def __init__(self):
        self.properties = []
        self.data = TextData()
        self.more_options = TextMoreOptions()
        self.masked_path = MaskedPath()

    def add_keyframe(self, time, item):
        self.data.keyframes.append(TextDataKeyframe(time, item))

    def get_value(self, time):
        return self.data.get_value(time)


## @ingroup Lottie
class FontPathOrigin(LottieEnum):
    Local = 0
    CssUrl = 1
    ScriptUrl = 2
    FontUrl = 3


## @ingroup Lottie
class Font(LottieObject):
    _props = [
        LottieProp("ascent", "ascent", float),
        LottieProp("font_family", "fFamily", str),
        LottieProp("name", "fName", str),
        LottieProp("font_style", "fStyle", str),
        LottieProp("path", "fPath", str),
        LottieProp("weight", "fWeight", str),
        LottieProp("origin", "origin", FontPathOrigin),
        LottieProp("css_class", "fClass", str),
    ]

    def __init__(self, font_family="sans", font_style="Regular", name=None):
        self.ascent = None
        self.font_family = font_family
        self.font_style = font_style
        self.name = name or "%s-%s" % (font_family, font_style)
        self.path = None
        self.weight = None
        self.origin = None
        self.css_class = None


## @ingroup Lottie
class FontList(LottieObject):
    _props = [
        LottieProp("list", "list", Font, True),
    ]

    def __init__(self):
        self.list = []

    def append(self, font):
        self.list.append(font)


## @ingroup Lottie
class CharacterData(LottieObject):
    """!
    Character shapes
    """
    _props = [
        LottieProp("shapes", "shapes", ShapeElement, True),
    ]

    def __init__(self):
        self.shapes = []


## @ingroup Lottie
class Chars(LottieObject):
    """!
    Defines character shapes to avoid loading system fonts
    """
    _props = [
        LottieProp("character", "ch", str, False),
        LottieProp("font_family", "fFamily", str, False),
        LottieProp("font_size", "size", float, False),
        LottieProp("font_style", "style", str, False),
        LottieProp("width", "w", float, False),
        LottieProp("data", "data", CharacterData, False),
    ]

    def __init__(self):
        ## Character Value
        self.character = ""
        ## Character Font Family
        self.font_family = ""
        ## Character Font Size
        self.font_size = 0
        ## Character Font Style
        self.font_style = "" # Regular
        ## Character Width
        self.width = 0
        ## Character Data
        self.data = CharacterData()

    @property
    def shapes(self):
        return self.data.shapes



from .. import objects


class RestructuredLayer:
    def __init__(self, lottie):
        self.lottie = lottie
        self.children_pre = []
        self.children_post = []
        self.structured = False
        self.shapegroup = None
        self.matte_target = False
        self.matte_source = None
        self.matte_id = None

    def add(self, child):
        c = self.children_pre if self.structured else self.children_post
        c.insert(0, child)


class RestructuredShapeGroup:
    def __init__(self, lottie):
        self.lottie = lottie
        self.children = []
        self.fill = None
        self.stroke = None
        self.layer = False
        self.paths = None
        self.stroke_above = False

    def empty(self):
        return not self.children

    def finalize(self, thresh=6):
        for g in self.subgroups:
            if g.layer:
                self.layer = True
                for gg in self.subgroups:
                    gg.layer = True
                return
        nchild = len(self.children)
        self.layer = nchild > thresh and self.lottie.name

    @property
    def subgroups(self):
        for g in self.children:
            if isinstance(g, RestructuredShapeGroup):
                yield g

    def add(self, child):
        self.children.insert(0, child)


class RestructuredModifier:
    def __init__(self, lottie, child):
        self.child = child
        self.lottie = lottie


class RestructuredPathMerger:
    def __init__(self):
        self.paths = []

    def append(self, path):
        self.paths.append(path)


class RestructuredAnimation:
    def __init__(self):
        self.layers = []
        self.precomp = {}


class AbstractBuilder:
    merge_paths = False

    def _on_animation(self, animation):
        raise NotImplementedError()

    def _on_shapegroup(self, shapegroup, out_parent):
        raise NotImplementedError()

    def _on_shape(self, shape, shapegroup, out_parent):
        raise NotImplementedError()

    def _on_merged_path(self, shape, shapegroup, out_parent):
        raise NotImplementedError()

    def _on_shape_modifier(self, shape, shapegroup, out_parent):
        raise NotImplementedError()

    def process(self, animation: objects.Animation):
        out_parent = self._on_animation(animation)

        restructured = self.restructure_animation(animation, self.merge_paths)
        for id, layers in restructured.precomp.items():
            self._on_precomp(id, out_parent, layers)

        for asset in animation.assets or []:
            self._on_asset(asset)

        if animation.fonts and animation.fonts.list:
            for font in animation.fonts.list:
                self._on_font(font)

        for layer_builder in restructured.layers:
            self.process_layer(layer_builder, out_parent)

    def _on_layer(self, layer_builder, out_parent):
        raise NotImplementedError()

    def _on_precomp(self, id, out_parent, layers):
        raise NotImplementedError()

    def _on_asset(self, asset):
        pass

    def _on_font(self, font):
        pass

    def process_layer(self, layer_builder, out_parent):
        out_layer = self._on_layer(layer_builder, out_parent)

        if out_layer is None:
            return

        for c in layer_builder.children_pre:
            self.process_layer(c, out_layer)

        shapegroup = getattr(layer_builder, "shapegroup", None)
        if shapegroup:
            self.shapegroup_process_children(shapegroup, out_layer)

        for c in layer_builder.children_post:
            self.process_layer(c, out_layer)

        self._on_layer_end(out_layer)

    def _on_layer_end(self, out_layer):
        pass

    def shapegroup_process_child(self, shape, shapegroup, out_parent):
        if isinstance(shape, RestructuredShapeGroup):
            return self._on_shapegroup(shape, out_parent)
        elif isinstance(shape, RestructuredPathMerger):
            return self._on_merged_path(shape, shapegroup, out_parent)
        elif isinstance(shape, RestructuredModifier):
            return self._on_shape_modifier(shape, shapegroup, out_parent)
        else:
            return self._on_shape(shape, shapegroup, out_parent)

    def shapegroup_process_children(self, shapegroup, out_parent):
        for shape in shapegroup.children:
            self.shapegroup_process_child(shape, shapegroup, out_parent)

    def restructure_animation(self, animation, merge_paths):
        restr = RestructuredAnimation()
        restr.layers = self.restructure_layer_list(animation.layers, merge_paths)
        if animation.assets:
            for asset in animation.assets:
                if isinstance(asset, objects.Precomp):
                    restr.precomp[asset.id] = self.restructure_layer_list(asset.layers, merge_paths)
        return restr

    def restructure_layer_list(self, layer_list, merge_paths):
        layers = {}
        flat_layers = []
        prev = None
        for layer in layer_list:
            laybuilder = RestructuredLayer(layer)
            flat_layers.append(laybuilder)

            if layer.index is not None:
                layers[layer.index] = laybuilder

            if isinstance(layer, objects.ShapeLayer):
                laybuilder.shapegroup = RestructuredShapeGroup(layer)
                laybuilder.layer = True
                for shape in layer.shapes:
                    self.restructure_shapegroup(shape, laybuilder.shapegroup, merge_paths)
                laybuilder.shapegroup.finalize()

            if layer.matte_mode not in {None, objects.MatteMode.Normal}:
                laybuilder.matte_source = prev
                if prev:
                    prev.matte_target = laybuilder

            prev = laybuilder

        top_layers = []
        for layer in flat_layers:
            layer.structured = True
            if layer.lottie.parent_index is not None:
                layers[layer.lottie.parent_index].add(layer)
            else:
                top_layers.insert(0, layer)

        return top_layers

    def restructure_shapegroup(self, shape, shape_group, merge_paths):
        if isinstance(shape, (objects.Fill, objects.GradientFill)):
            if not shape_group.fill:
                shape_group.fill = shape
        elif isinstance(shape, objects.BaseStroke):
            if not shape_group.stroke or shape_group.stroke.width.get_value(0) < shape.width.get_value(0):
                shape_group.stroke = shape
                if not shape_group.fill:
                    shape_group.stroke_above = True
        elif isinstance(shape, (objects.Path)):
            if merge_paths:
                if not shape_group.paths:
                    shape_group.paths = RestructuredPathMerger()
                    shape_group.add(shape_group.paths)
                shape_group.paths.append(shape)
            else:
                shape_group.add(shape)
        elif isinstance(shape, (objects.Group)):
            subgroup = RestructuredShapeGroup(shape)
            shape_group.add(subgroup)
            merge_paths = self.merge_paths and not any(isinstance(p, objects.Group) for p in shape.shapes)
            for subshape in shape.shapes:
                self.restructure_shapegroup(subshape, subgroup, merge_paths)
            subgroup.finalize()
        elif isinstance(shape, (objects.Modifier)):
            if shape_group.children:
                ch = shape_group.children.pop(0)
                shape_group.add(RestructuredModifier(shape, ch))
        elif isinstance(shape, (objects.ShapeElement)):
            shape_group.add(shape)
        elif isinstance(shape, (objects.base.CustomObject)):
            if self._custom_object_supported(shape):
                shape_group.add(shape)
            else:
                self.restructure_shapegroup(shape.wrapped, shape_group, self.merge_paths)

    def _custom_object_supported(self, shape):
        return False



from ..parsers.baseporter import Baseporter, Loader


class ImporterLoader(Loader):
    def __init__(self):
        super().__init__(__file__, __name__, "import")

    @property
    def importers(self):
        return self.items

    def set_options(self, parser):
        group = parser.add_argument_group("Generic input options")

        super().set_options(parser)

        return group


importers = ImporterLoader()
importer = importers.decorator



#!/usr/bin/env python3

import json
import sys
import os
import argparse
sys.path.insert(0, os.path.join(
    os.path.dirname(os.path.dirname(os.path.abspath(__file__))),
    "lib"
))
from lottie.utils.font import fonts
from lottie import __version__

parser = argparse.ArgumentParser(
    description="List available fonts",
)
parser.add_argument("--version", "-v", action="version", version="%(prog)s - python-lottie " + __version__)
parser.add_argument(
    "fonts",
    default=[],
    nargs="*",
    help="Font family names"
)


def _fonts(flist):
    for font in flist:
        if font in fonts:
            yield font


if __name__ == "__main__":
    ns = parser.parse_args()
    font_iter = _fonts(ns.fonts) if ns.fonts else fonts
    for font in font_iter:
        print("* %s" % font)
        for style in fonts[font].files.keys():
            print("  * %s" % " ".join(style))



from .base import LottieObject, Index, LottieProp
from .layers import Layer


## @ingroup Lottie
class Composition(LottieObject):
    """!
    Base class for layer holders
    """
    _props = [
        LottieProp("layers", "layers", Layer, True),
    ]

    def __init__(self):
        super().__init__()
        ## List of Composition Layers
        self.layers = [] # ShapeLayer, SolidLayer, CompLayer, ImageLayer, NullLayer, TextLayer

        self._index_gen = Index()

    def layer(self, index):
        for layer in self.layers:
            if layer.index == index:
                return layer
        raise IndexError("No layer %s" % index)

    def add_layer(self, layer: Layer):
        """!
        @brief Appends a layer to the composition
        @see insert_layer
        """
        return self.insert_layer(len(self.layers), layer)

    @classmethod
    def load(cls, lottiedict):
        obj = super().load(lottiedict)
        obj._fixup()
        return obj

    def _fixup(self):
        for layer in self.layers:
            layer.composition = self

    def insert_layer(self, index, layer: Layer):
        """!
        @brief Inserts a layer to the composition
        @note Layers added first will be rendered on top of later layers
        """
        self.layers.insert(index, layer)
        self.prepare_layer(layer)
        return layer

    def prepare_layer(self, layer: Layer):
        layer.composition = self
        if layer.index is None:
            layer.index = next(self._index_gen)
        self._on_prepare_layer(layer)

    def _on_prepare_layer(self, layer):
        raise NotImplementedError

    def clone(self):
        c = super().clone()
        c._index_gen._i = self._index_gen._i
        return c

    def remove_layer(self, layer: Layer):
        """!
        @brief Removes a layer (and all of its children) from this composition
        @param layer    Layer to be removed
        """
        if layer.composition is not self:
            return

        children = list(layer.children)

        layer.composition = None
        self.layers.remove(layer)

        for c in children:
            self.remove_layer(c)



import enum
import inspect
import importlib
from ..nvector import NVector
from ..utils.color import Color, ColorMode


class LottieBase:
    """!
    Base class for Lottie JSON objects bindings
    """
    def to_dict(self):
        """!
        Serializes into a JSON object fit for the Lottie format
        """
        raise NotImplementedError

    @classmethod
    def load(cls, lottiedict):
        """!
        Loads from a JSON object
        @returns An instance of the class
        """
        raise NotImplementedError

    def clone(self):
        """!
        Returns a copy of the object
        """
        raise NotImplementedError


class EnumMeta(enum.EnumMeta):
    """!
    Hack to counter-hack the hack in enum meta
    """
    def __new__(cls, name, bases, classdict):
        classdict["__reduce_ex__"] = lambda *a, **kw: None  # pragma: no cover
        return super().__new__(cls, name, bases, classdict)


class LottieEnum(LottieBase, enum.Enum, metaclass=EnumMeta):
    """!
    Base class for enum-like types in the Lottie JSON structure
    """
    def to_dict(self):
        return self.value

    @classmethod
    def load(cls, lottieint):
        return cls(lottieint)

    def clone(self):
        return self


class PseudoList:
    """!
    List tag for some weird values in the Lottie JSON
    """
    pass


class LottieValueConverter:
    """!
    Factory for property types that require special conversions
    """
    def __init__(self, py, lottie, name=None):
        self.py = py
        self.lottie = lottie
        self.name = name or "%s but displayed as %s" % (self.py.__name__, self.lottie.__name__)

    def py_to_lottie(self, val):
        return self.lottie(val)

    def lottie_to_py(self, val):
        return self.py(val)

    @property
    def __name__(self):
        return self.name


## For values in Lottie that are bools but ints in the JSON
PseudoBool = LottieValueConverter(bool, int, "0-1 int")


class LottieProp:
    """!
    Lottie <-> Python property mapper
    """
    def __init__(self, name, lottie, type=float, list=False, cond=None):
        ## Name of the Python property
        self.name = name
        ## Name of the Lottie JSON property
        self.lottie = lottie
        ## Type of the property
        ## @see LottieValueConverter, PseudoBool
        self.type = type
        ## Whether the property is a list of self.type
        ## @see PseudoList
        self.list = list
        ## Condition on when the property is loaded from the Lottie JSON
        self.cond = cond

    def get(self, obj):
        """!
        Returns the value of the property from a Python object
        """
        return getattr(obj, self.name)

    def set(self, obj, value):
        """!
        Sets the value of the property from a Python object
        """
        if isinstance(getattr(obj.__class__, self.name, None), property):
            return
        return setattr(obj, self.name, value)

    def load_from_parent(self, lottiedict):
        """!
        Returns the value for this property from a JSON dict representing the parent object
        @returns The loaded value or @c None if the property is not in @p lottiedict
        """
        if self.lottie in lottiedict:
            return self.load(lottiedict[self.lottie])
        return None

    def load_into(self, lottiedict, obj):
        """!
        Loads from a Lottie dict into an object
        """
        if self.cond and not self.cond(lottiedict):
            return
        self.set(obj, self.load_from_parent(lottiedict))

    def load(self, lottieval):
        """!
        Loads the property from a JSON value
        @returns the Python equivalent of the JSON value
        """
        if self.list is PseudoList and isinstance(lottieval, list):
            return self._load_scalar(lottieval[0])
        elif self.list is True:
            return list(filter(lambda x: x is not None, (
                self._load_scalar(it)
                for it in lottieval
            )))
        return self._load_scalar(lottieval)

    def _load_scalar(self, lottieval):
        if lottieval is None:
            return None
        if inspect.isclass(self.type) and issubclass(self.type, LottieBase):
            return self.type.load(lottieval)
        elif isinstance(self.type, type) and isinstance(lottieval, self.type):
            return lottieval
        elif isinstance(self.type, LottieValueConverter):
            return self.type.lottie_to_py(lottieval)
        elif self.type is NVector:
            return NVector(*lottieval)
        elif self.type is Color:
            return Color(*lottieval)
        if isinstance(lottieval, list) and lottieval:
            lottieval = lottieval[0]
        return self.type(lottieval)

    def to_dict(self, obj):
        """!
        Converts the value of the property as from @p obj into a JSON value
        @param obj LottieObject with this property
        """
        val = self.get(obj)
        if isinstance(self.type, LottieValueConverter):
            val = self.type.py_to_lottie(val)

        val = self._basic_to_dict(val)

        if self.list is PseudoList:
            if not isinstance(obj, list):
                return [val]
        return val

    def _basic_to_dict(self, v):
        if isinstance(v, LottieBase):
            return v.to_dict()
        elif isinstance(v, Color):
            return list(map(self._basic_to_dict, v.to_rgb().components))
        elif isinstance(v, NVector):
            return list(map(self._basic_to_dict, v.components))
        elif isinstance(v, (list, tuple)):
            return list(map(self._basic_to_dict, v))
        elif isinstance(v, (int, str, bool)):
            return v
        elif isinstance(v, float):
            if v % 1 == 0:
                return int(v)
            return v # round(v, 3)
        else:
            raise Exception("Unknown value {!r}".format(v))

    def __repr__(self):
        return "<LottieProp %s:%s>" % (self.name, self.lottie)

    def clone_value(self, value):
        if isinstance(value, (list, tuple)):
            return [self.clone_value(v) for v in value]
        if isinstance(value, (LottieBase, NVector)):
            return value.clone()
        if isinstance(value, (int, float, bool, str)) or value is None:
            return value
        raise Exception("Could not convert {!r}".format(value))


class LottieObjectMeta(type):
    def __new__(cls, name, bases, attr):
        props = []
        for base in bases:
            if type(base) == cls:
                props += base._props
        attr["_props"] = props + attr.get("_props", [])
        return super().__new__(cls, name, bases, attr)


class LottieObject(LottieBase, metaclass=LottieObjectMeta):
    """!
    @brief Base class for mapping Python classes into Lottie JSON objects
    """
    def __init__(self):
        pass

    def to_dict(self):
        return {
            prop.lottie: prop.to_dict(self)
            for prop in self._props
            if prop.get(self) is not None
        }

    @classmethod
    def load(cls, lottiedict):
        if "__pyclass" in lottiedict:
            return CustomObject.load(lottiedict)
        if not lottiedict:
            return None
        cls = cls._load_get_class(lottiedict)
        obj = cls()
        for prop in cls._props:
            prop.load_into(lottiedict, obj)
        return obj

    @classmethod
    def _load_get_class(cls, lottiedict):
        return cls

    def find(self, search, propname="name"):
        """!
        @param search   The value of the property to search
        @param propname The name of the property used to search
        @brief Recursively searches for child objects with a matching property
        """
        if getattr(self, propname, None) == search:
            return self
        for prop in self._props:
            v = prop.get(self)
            if isinstance(v, LottieObject):
                found = v.find(search, propname)
                if found:
                    return found
            elif isinstance(v, list) and v and isinstance(v[0], LottieObject):
                for obj in v:
                    found = obj.find(search, propname)
                    if found:
                        return found
        return None

    def find_all(self, type, predicate=None, include_self=True):
        """!
        Find all child objects that match a predicate
        @param type         Type (or tuple of types) of the objects to match
        @param predicate    Function that returns true on the objects to find
        @param include_self Whether should counsider `self` for a potential match
        """

        if isinstance(self, type) and include_self:
            if not predicate or predicate(self):
                yield self

        for prop in self._props:
            v = prop.get(self)

            if isinstance(v, LottieObject):
                for found in v.find_all(type, predicate, True):
                    yield found
            elif isinstance(v, list) and v and isinstance(v[0], LottieObject):
                for child in v:
                    for found in child.find_all(type, predicate, True):
                        yield found

    def clone(self):
        obj = self.__class__()
        for prop in self._props:
            v = prop.get(self)
            prop.set(obj, prop.clone_value(v))
        return obj

    def clone_into(self, other):
        for prop in self._props:
            v = prop.get(self)
            prop.set(other, prop.clone_value(v))

    def __str__(self):
        return type(self).__name__


class Index:
    """!
    @brief Simple iterator to generate increasing integers
    """
    def __init__(self):
        self._i = -1

    def __next__(self):
        self._i += 1
        return self._i


class CustomObject(LottieObject):
    """!
    Allows extending the Lottie shapes with custom Python classes
    """
    wrapped_lottie = LottieObject

    def __init__(self):
        self.wrapped = self.wrapped_lottie()

    @classmethod
    def load(cls, lottiedict):
        ld = lottiedict.copy()
        classname = ld.pop("__pyclass")
        modn, clsn = classname.rsplit(".", 1)
        subcls = getattr(importlib.import_module(modn), clsn)
        obj = subcls()
        for prop in subcls._props:
            prop.load_into(lottiedict, obj)
        obj.wrapped = subcls.wrapped_lottie.load(ld)
        return obj

    def clone(self):
        obj = self.__class__(**self.to_pyctor())
        obj.wrapped = self.wrapped.clone()
        return obj

    def to_dict(self):
        dict = self.wrapped.to_dict()
        dict["__pyclass"] = "{0.__module__}.{0.__name__}".format(self.__class__)
        dict.update(LottieObject.to_dict(self))
        return dict

    def _build_wrapped(self):
        return self.wrapped_lottie()

    def refresh(self):
        self.wrapped = self._build_wrapped()


class ObjectVisitor:
    DONT_RECURSE = object()

    def __call__(self, lottie_object):
        self._process(lottie_object)

    def _process(self, lottie_object):
        self.visit(lottie_object)
        for p in lottie_object._props:
            pval = p.get(lottie_object)
            if self.visit_property(lottie_object, p, pval) is not self.DONT_RECURSE:
                if isinstance(pval, LottieObject):
                    self._process(pval)
                elif isinstance(pval, list) and pval and isinstance(pval[0], LottieObject):
                    for c in pval:
                        self._process(c)

    def visit(self, object):
        pass

    def visit_property(self, object, property, value):
        pass



import enum


class FrameTime:
    class Unit(enum.Enum):
        Frame = "f"
        Seconds = "s"

    def __init__(self, value, unit):
        self.value = value
        self.unit = unit

    def __eq__(self, other):
        return self.value == other.value and self.unit == other.unit

    def __ne__(self, other):
        return self.value == other.value and self.unit == other.unit

    def __str__(self):
        return "%s%s" % (self.value, self.unit.value)

    def __repr__(self):
        return "<%s %s>" % (self.__class__.__name__, self)

    @classmethod
    def frame(cls, amount):
        return cls(amount, cls.Unit.Frame)

    @classmethod
    def seconds(cls, amount):
        return cls(amount, cls.Unit.Seconds)

    @classmethod
    def parse_string(cls, value_str, canvas):
        if " " in value_str:
            value = 0
            unit = cls.Unit.Frame
            for sub in value_str.split():
                sv = float(sub[:-1])
                if sub[-1] == "s":
                    sv *= canvas.fps
                value += sv
        else:
            value = float(value_str[:-1])
            unit = cls.Unit(value_str[-1])
        return FrameTime(value, unit)



import warnings
from .base import LottieObject, LottieProp, PseudoBool, LottieEnum, LottieValueConverter
from .effects import Effect
from .helpers import Transform, Mask, VisualObject, BlendMode
from .shapes import ShapeElement
from .text import TextAnimatorData
from .properties import Value, MultiDimensional
from ..utils.color import Color, color_from_hex, color_to_hex
from .styles import LayerStyle


## @ingroup Lottie
## @todo SVG masks
class MatteMode(LottieEnum):
    Normal = 0
    Alpha = 1
    InvertedAlpha = 2
    Luma = 3
    InvertedLuma = 4


#ingroup Lottie
class Layer(VisualObject):
    """!
    Base class for all layers
    """
    _props = [
        LottieProp("threedimensional", "ddd", PseudoBool, False),
        LottieProp("hidden", "hd", bool, False),
        LottieProp("type", "ty", int, False),
        LottieProp("index", "ind", int, False),
        LottieProp("parent_index", "parent", int, False),
        LottieProp("time_stretch", "sr", float, False),
        LottieProp("in_point", "ip", float, False),
        LottieProp("out_point", "op", float, False),
        LottieProp("start_time", "st", float, False),
    ]
    ## %Layer type.
    ## @see https://github.com/bodymovin/bodymovin-extension/blob/master/bundle/jsx/enums/layerTypes.jsx
    type = None
    _classses = {}

    def __init__(self):
        super().__init__()

        ## Whether the layer is threedimensional
        self.threedimensional = 0
        ## Whether the layer is hidden
        self.hidden = None
        ## Index that can be used for parenting and referenced in expressions
        self.index = None
        ## Must be the `ind` property of another layer
        self.parent_index = None
        self.time_stretch = 1
        ## Frame when the layer becomes visible
        self.in_point = None
        ## Frame when the layer becomes invisible
        self.out_point = None
        ## Start Time of layer. Sets the start time of the layer.
        self.start_time = 0

    @classmethod
    def _load_get_class(cls, lottiedict):
        if not Layer._classses:
            Layer._classses = {
                sc.type: sc
                for sc in Layer.__subclasses__() + VisualLayer.__subclasses__()
                if sc.type is not None
            }
        type_id = lottiedict["ty"]
        if type_id not in Layer._classses:
            warnings.warn("Unknown layer type: %s" % type_id)
            return Layer
        return Layer._classses[type_id]


## @ingroup Lottie
class VisualLayer(Layer):
    """!
    Base class for layers that have a visual component
    """
    _props = [
        LottieProp("collapse_transform", "cp", bool, False),
        LottieProp("transform", "ks", Transform, False),
        LottieProp("auto_orient", "ao", PseudoBool, False),

        LottieProp("blend_mode", "bm", BlendMode, False),

        LottieProp("matte_mode", "tt", MatteMode, False),
        LottieProp("css_class", "cl", str, False),
        LottieProp("layer_xml_id", "ln", str, False),
        LottieProp("layer_xml_tag_name", "tg", str, False),

        LottieProp("motion_blur", "mb", bool, False),
        LottieProp("layer_style", "sy", LayerStyle, True),

        LottieProp("has_masks", "hasMask", bool, False),
        LottieProp("masks", "masksProperties", Mask, True),
        LottieProp("effects", "ef", Effect, True),
        LottieProp("matte_target", "td", int, False),
    ]

    @property
    def has_masks(self):
        """!
        Whether the layer has some masks applied
        """
        return bool(self.masks) if getattr(self, "masks") is not None else None

    def __init__(self):
        super().__init__()

        self.collapse_transform = None
        ## Transform properties
        self.transform = Transform()
        ## Auto-Orient along path AE property.
        self.auto_orient = False

        ## CSS class used by the SVG renderer
        self.css_class = None
        ## `id` attribute used by the SVG renderer
        self.layer_xml_id = None
        ## tag name used by the SVG renderer
        self.layer_xml_tag_name = None

        ## Whether motion blur is enabled for the layer
        self.motion_blur = None
        ## Styling effects for this layer
        self.layer_style = None

        ## List of Effects
        self.effects = None
        ## Layer Time Stretching
        self.stretch = 1
        ## List of Masks
        self.masks = None
        ## Blend Mode
        self.blend_mode = BlendMode.Normal
        ## Matte mode, the layer will inherit the transparency from the layer above
        self.matte_mode = None
        self.matte_target = None
        ## Composition owning the layer, set by add_layer
        self.composition = None

    def add_child(self, layer):
        if not self.composition or self.index is None:
            raise Exception("Must set composition / index first")
        self._child_inout_auto(layer)
        self.composition.add_layer(layer)
        layer.parent_index = self.index
        return layer

    def _child_inout_auto(self, layer):
        if layer.in_point is None:
            layer.in_point = self.in_point
        if layer.out_point is None:
            layer.out_point = self.out_point

    @property
    def parent(self):
        if self.parent_index is None:
            return None
        return self.composition.layer(self.parent_index)

    @parent.setter
    def parent(self, layer):
        if layer is None:
            self.parent_index = None
        else:
            self.parent_index = layer.index
            layer._child_inout_auto(self)

    @property
    def children(self):
        for layer in self.composition.layers:
            if layer.parent_index == self.index:
                yield layer

    def __repr__(self):
        return "<%s %s %s>" % (type(self).__name__, self.index, self.name)

    def __str__(self):
        return "%s %s" % (
            self.name or super().__str__(),
            self.index if self.index is not None else ""
        )

    def remove(self):
        """!
        @brief Removes this layer from the componsitin
        """
        self.composition.remove_layer(self)


## @ingroup Lottie
class NullLayer(VisualLayer):
    """!
    Layer with no data, useful to group layers together
    """
    ## %Layer type.
    type = 3

    def __init__(self):
        super().__init__()


## @ingroup Lottie
class TextLayer(VisualLayer):
    _props = [
        LottieProp("data", "t", TextAnimatorData, False),
    ]
    ## %Layer type.
    type = 5

    def __init__(self):
        super().__init__()
        ## Text Data
        self.data = TextAnimatorData()


## @ingroup Lottie
class ShapeLayer(VisualLayer):
    """!
    Layer containing ShapeElement objects
    """
    _props = [
        LottieProp("shapes", "shapes", ShapeElement, True),
    ]
    ## %Layer type.
    type = 4

    def __init__(self):
        super().__init__()
        ## Shape list of items
        self.shapes = [] # ShapeElement

    def add_shape(self, shape):
        self.shapes.append(shape)
        return shape

    def insert_shape(self, index, shape):
        self.shapes.insert(index, shape)
        return shape


## @ingroup Lottie
## @todo SIF I/O
class ImageLayer(VisualLayer):
    _props = [
        LottieProp("image_id", "refId", str, False),
    ]
    ## %Layer type.
    type = 2

    def __init__(self, image_id=""):
        super().__init__()
        ## id pointing to the source image defined on 'assets' object
        self.image_id = image_id


## @ingroup Lottie
class PreCompLayer(VisualLayer):
    _props = [
        LottieProp("reference_id", "refId", str, False),
        LottieProp("time_remapping", "tm", Value, False),
        LottieProp("width", "w", int, False),
        LottieProp("height", "h", int, False),
    ]
    ## %Layer type.
    type = 0

    def __init__(self, reference_id=""):
        super().__init__()
        ## id pointing to the source composition defined on 'assets' object
        self.reference_id = reference_id
        ## Comp's Time remapping
        self.time_remapping = None
        ## Width
        self.width = 512
        ## Height
        self.height = 512


ColorString = LottieValueConverter(color_from_hex, color_to_hex, "Color string")


## @ingroup Lottie
class SolidColorLayer(VisualLayer):
    """!
    Layer with a solid color rectangle
    """
    _props = [
        LottieProp("color", "sc", ColorString, False),
        LottieProp("height", "sh", float, False),
        LottieProp("width", "sw", float, False),
    ]
    ## %Layer type.
    type = 1

    def __init__(self, color=Color(), width=512, height=512):
        super().__init__()
        ## Color of the layer as a @c \#rrggbb hex
        self.color = color
        ## Height of the layer.
        self.height = height
        ## Width of the layer.
        self.width = width


#ingroup Lottie
class CameraLayer(Layer):
    """!
    3D Camera
    """
    type = 13
    _props = [
        LottieProp("type", "ty", int, False),
        LottieProp("transform", "ks", Transform, False),
        LottieProp("perspective", "pe", Value, False),
    ]

    def __init__(self):
        super().__init__()

        ## Layer transform
        self.transform = None
        ## Distance from the Z=0 plane.
        ## Small values yield a higher perspective effect.
        self.perspective = None


#ingroup Lottie
class DataLayer(Layer):
    _props = [
        LottieProp("type", "ty", int, False),
        LottieProp("data_source_id", "refId", str, False),
    ]
    type = 15

    def __init__(self):
        super().__init__()

        ## ID of the data source in assets
        self.data_source_id = None


#ingroup Lottie
class AudioSettings(LottieObject):
    _props = [
        LottieProp("level", "lv", MultiDimensional, False),
    ]

    def __init__(self):
        super().__init__()

        self.level = MultiDimensional([0, 0])


#ingroup Lottie
class AudioLayer(Layer):
    """!
    A layer playing sounds
    """
    _props = [
        LottieProp("type", "ty", int, False),
        LottieProp("audio_settings", "au", AudioSettings, False),
        LottieProp("sound_id", "refId", str, False),
    ]
    type = 6

    def __init__(self):
        super().__init__()

        self.audio_settings = None
        ## ID of the sound as specified in the assets
        self.sound_id = None



import re

from .font import FontStyle
from ..nvector import NVector
from ..parsers.svg.importer import parse_color
from ..objects.shapes import Group, Fill


class FancyStyle:
    def __init__(self, font: FontStyle, color: NVector, font_size: float, offset: NVector, scale: NVector, rotation: float):
        self.font = font
        self.color = color.clone()
        self.font_size = font_size
        self.offset = offset.clone()
        self.scale = scale.clone()
        self.rotation = rotation

    def clone(self):
        return FancyStyle(self.font, self.color, self.font_size, self.offset, self.scale, self.rotation)

    def render(self, text, pos, start_x):
        pos += self.offset
        g = self.font.renderer.render(text, self.font_size, pos, True, start_x)
        g.add_shape(Fill(self.color))
        if self.scale.x != 1 or self.scale.y != 1 or self.rotation != 0:
            center = g.bounding_box().center()
            g.transform.anchor_point.value = center
            g.transform.position.value += center
            g.transform.scale.value = self.scale * 100
            g.transform.rotation.value = self.rotation
        pos -= self.offset
        return g


class FancyTextRenderer:
    _regex = re.compile(r'\\([a-z0-9]+)(?:\{([^}]*)\})?')

    def __init__(self, font: FontStyle, default_color: NVector, font_size: float):
        self.font = font
        self.font_size = font_size
        self.default_color = default_color
        self.groups = []

    @staticmethod
    def _int_arg(value, default=0):
        try:
            return float(value)
        except (ValueError, TypeError):
            return default

    def render(self, text: str, pos: NVector = None):
        if pos is None:
            pos = NVector(0, 0)

        line_start = pos.x
        default_style = FancyStyle(self.font, self.default_color, self.font_size, NVector(0, 0), NVector(1, 1), 0)
        style = default_style.clone()
        container = Group()
        last_pos = 0

        for match in self._regex.finditer(text):
            prev_text = text[last_pos:match.start()]
            last_pos = match.end()
            if prev_text:
                container.insert_shape(0, style.render(prev_text, pos, line_start))

            style = style.clone()

            command = match.group(1)
            arg = match.group(2)

            if command == "color":
                if arg:
                    style.color = parse_color(arg, self.default_color)
                else:
                    style.color = self.default_color
            elif command == "huge":
                style.font_size = self.font_size * 2
            elif command == "large":
                style.font_size = self.font_size * 1.5
            elif command == "normal":
                style.font_size = self.font_size
            elif command == "small":
                style.font_size = self.font_size / 1.5
            elif command == "tiny":
                style.font_size = self.font_size / 2
            elif command == "super":
                style.offset.y -= self.font_size / 2
            elif command == "sub":
                style.offset.y += self.font_size / 2
            elif command == "center":
                style.offset.y = 0
            elif command == "clear":
                style = default_style.clone()
            elif command == "flip":
                style.scale.x *= -1
            elif command == "vflip":
                style.scale.y *= -1
            elif command == "r":
                pos.x = line_start
            elif command == "n":
                pos.x = line_start
                pos.y += self.font.line_height
            elif command == "hspace":
                pos.x += self._int_arg(arg)
            elif command == "rot":
                style.rotation = self._int_arg(arg)

            # Eat space after no argument command
            if arg is None and len(text) > last_pos and text[last_pos] == ' ':
                last_pos += 1

        last_text = text[last_pos:]
        if last_text:
            container.insert_shape(0, style.render(last_text, pos, line_start))

        if len(container.shapes) > 1:
            container.next_x = container.shapes[-2].next_x
        else:
            container.next_x = pos.x

        return container


def render_fancy_text(text: str, font: FontStyle, default_color: NVector, font_size: float, pos: NVector = None):
    renderer = FancyTextRenderer(font, default_color, font_size)
    return renderer.render(text, pos)



from .base import importer
from ..parsers.baseporter import ExtraOption
from ..parsers.svg import parse_svg_file
from ..parsers.tgs import open_maybe_gzipped


@importer("SVG", ["svg", "svgz"], [
    ExtraOption(
        "layer_frames", type=int, default=0,
        help="If greater than 0, treats every layer in the SVG as a different animation frame,\n"
        "greater values increase the time each frames lasts for."),
    ExtraOption("n_frames", type=int, default=60),
    ExtraOption("framerate", type=int, default=60),
])
def import_svg(file, *a, **kw):
    return open_maybe_gzipped(file, lambda svgfile: parse_svg_file(svgfile, *a, **kw))



from PyQt5.QtGui import QIcon
from PyQt5.QtCore import Qt, QTimer
from PyQt5.QtCore import pyqtSignal as Signal
from PyQt5.QtCore import pyqtSlot as Slot
from PyQt5.QtWidgets import *


class TimelineWidget(QWidget):
    frame_changed = Signal(int)

    def __init__(self):
        super().__init__()
        layout_slider = QHBoxLayout()
        layout_slider.setContentsMargins(0, 0, 0, 0)
        self.setLayout(layout_slider)
        self.button_play = QPushButton()
        self.button_play.setCheckable(True)
        self.button_play.toggled.connect(self.play_toggle)
        layout_slider.addWidget(self.button_play)
        self.slider = QSlider(Qt.Horizontal)
        layout_slider.addWidget(self.slider)
        self.slider_spin = QSpinBox()
        layout_slider.addWidget(self.slider_spin)
        self.slider.valueChanged.connect(self.slider_spin.setValue)
        self.slider_spin.valueChanged.connect(self.slider.setValue)
        self.slider.valueChanged.connect(self.frame_changed)

        self.framerate = 60
        self.timer = QTimer(self)
        self.timer.timeout.connect(self._next_frame)
        self.stop()

        self._min = 0
        self._max = 99

    def set_min_max(self, min, max):
        self._min = min
        self._max = max
        self.slider.setMinimum(min)
        self.slider.setMaximum(max)
        self.slider_spin.setMinimum(min)
        self.slider_spin.setMaximum(max)

    @Slot(int)
    def set_frame(self, frame):
        self.slider.setValue(frame)

    @property
    def frame(self):
        return self.slider.value()

    def _next_frame(self):
        nf = self.frame + 1
        if nf > self._max:
            nf = self._min
        self.slider.setValue(nf)

    @Slot()
    def stop(self):
        self.timer.stop()
        self.button_play.setIcon(QIcon.fromTheme("media-playback-start"))
        self.button_play.setText("Play")
        self.button_play.setChecked(False)

    @Slot()
    def play(self):
        if not self.isEnabled():
            self.stop()
            return

        self.button_play.setIcon(QIcon.fromTheme("media-playback-stop"))
        self.button_play.setText("Stop")

        self.timer.start(1000/self.fps)

    @Slot(bool)
    def play_toggle(self, play):
        if play:
            self.play()
        else:
            self.stop()
